///////////////////////////////////////////////////////////////////////////////
//Copyright (C) 2003 Jason Baldridge and University of Edinburgh (Michael White)
//Copyright (C) 2012 Christos Christodoulopoulos

//This library is free software; you can redistribute it and/or
//modify it under the terms of the GNU Lesser General Public
//License as published by the Free Software Foundation; either
//version 2.1 of the License, or (at your option) any later version.

//This library is distributed in the hope that it will be useful,
//but WITHOUT ANY WARRANTY; without even the implied warranty of
//MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//GNU Lesser General Public License for more details.

//You should have received a copy of the GNU Lesser General Public
//License along with this program; if not, write to the Free Software
//Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.


package edinburgh.ccg.ccgCheck.grammar;

import java.util.List;

import edinburgh.ccg.ccgCheck.utils.Category;
import edinburgh.ccg.ccgCheck.utils.unify.UnifyFailure;

/**
 * Implements some default behavior for Rule objects.
 *
 * @author  Jason Baldridge
 * @author  Michael White
 * @author 	Christos Christodoulopoulos
 */
public abstract class AbstractRule implements Rule {

	/** The name of this rule. */
	protected String _name;

	/** The rule group which contains this rule. */
	protected RuleGroup _ruleGroup;

	/**
	 * Applies the rule to the given input signs, adding to the given list of results.
	 */
	public void applyRule(Category[] inputs, List<Category> results, String expansion) {

		if (inputs.length != arity()) { // shouldn't happen
			throw new RuntimeException("Inputs must have length " + arity());
		}

		Category[] cats = new Category[inputs.length];
		for (int i=0; i < cats.length; i++) {
			cats[i] = inputs[i];
		}
		try {
			List<Category> resultCats = applyRule(cats);
			if (resultCats.isEmpty()) return;
			else results.addAll(resultCats);
			
		} catch (UnifyFailure uf) {}
	}

	/**
	 * The number of arguments this rule takes.  For example, the arity of the
	 * forward application rule of categorial grammar (X/Y Y => Y) is 2.
	 *
	 * @return the number of arguments this rule takes
	 **/
	public abstract int arity();

	/**
	 * Apply this rule to some input categories.
	 *
	 * @param inputs the input categories to try to combine
	 * @return the categories resulting from using this rule to combine the
	 *         inputs
	 * @exception UnifyFailure if the inputs cannot be combined by this rule
	 **/
	public abstract List<Category> applyRule(Category[] inputs) throws UnifyFailure;


	/** Prints an apply instance for the given categories to System.out. */
	protected void showApplyInstance(Category[] inputs) {
		StringBuffer sb = new StringBuffer();
		sb.append(_name).append(": ");

		for (int i=0; i < inputs.length; i++) {
			sb.append(inputs[i]).append(' ');
		}

		System.out.println(sb);
	}

	/** Prints an apply instance for the given categories to System.out. */
	protected void showApplyInstance(Category first, Category second) {
		Category[] ca = {first,second};
		showApplyInstance(ca);
	}


	/**
	 * Returns the name of this rule.
	 */
	public String name() {
		return _name;
	}

	/**
	 * Returns the rule group which contains this rule.
	 */
	public RuleGroup getRuleGroup() { return _ruleGroup; }

	/**
	 * Sets this rule's rule group.
	 */
	public void setRuleGroup(RuleGroup ruleGroup) { _ruleGroup = ruleGroup; }
}