///////////////////////////////////////////////////////////////////////////////
//Copyright (C) 2003-5 Jason Baldridge, Gann Bierner and 
//University of Edinburgh (Michael White)

//This library is free software; you can redistribute it and/or
//modify it under the terms of the GNU Lesser General Public
//License as published by the Free Software Foundation; either
//version 2.1 of the License, or (at your option) any later version.

//This library is distributed in the hope that it will be useful,
//but WITHOUT ANY WARRANTY; without even the implied warranty of
//MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//GNU Lesser General Public License for more details.

//You should have received a copy of the GNU Lesser General Public
//License along with this program; if not, write to the Free Software
//Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.


package edinburgh.ccg.ccgCheck.utils;

import edinburgh.ccg.ccgCheck.utils.unify.Substitution;
import edinburgh.ccg.ccgCheck.utils.unify.UnifyFailure;

/**
 * Abstract category adapter for CCG categories.
 *
 * @author      Gann Bierner
 * @author      Jason Baldridge
 * @author      Michael White
 * @version     $Revision: 1.15 $, $Date: 2006/09/29 06:55:40 $ 
 */
public abstract class AbstractCat implements Category { 

	private static final long serialVersionUID = 1L;

	/** The hash code, if already computed. */
	private int _hashCode = -1;

	/** Default constructor. */
	public AbstractCat() {}

	//-----------------------------------------------------------------
	// methods from Category which should be implemented by subclasses of
	// AbstractCat

	public abstract String toString();

	/**
	 * Returns the supertag for the category.
	 */
	public abstract String getSupertag();

	/**
	 * This will return a TeX formatted representation for a category.
	 * If toTeX() is not implemented for this category, the normal
	 * toString() method is invoked instead.
	 */
	public String toTeX() {
		return toString();
	}

	public abstract Category copy();

	public abstract Object fill (Substitution s) throws UnifyFailure;
	public abstract void unifyCheck (Object u) throws UnifyFailure;
	/** NB: The LF does not participate in unification. */
	public abstract Object unify (Object u, Substitution sub) 
	throws UnifyFailure;

	/**
	 * Returns the target category of this category.
	 */
	public abstract TargetCat getTarget();


	//-----------------------------------------------------------------
	// implemented methods from Category

	public boolean shallowEquals(Object o) { 
		return equals(o); 
	}

	public void forall(CategoryFcn f) { 
		f.forall(this); 
	}

	// methods to support printing of Categories
	public String prettyPrint() { return prettyPrint(""); }
	protected String prettyPrint(String pad) { return pad+toString(); }

	protected int prettyLength(String s) {
		int max=0, cur=0;
		for(int i=0; i<s.length(); i++)
			if(s.charAt(i) == '\n') {
				max = Math.max(cur, max);
				cur = 0;
			} else cur++;
		return Math.max(max, cur);
	}


	/** 
	 * Returns a hash code for this category. 
	 * The hash code handles equivalence up to variable names 
	 * as long as features and predicates are in the same order.
	 * The implementation calls hashCode(varMap) and 
	 * caches the result.
	 */
	public int hashCode() {
		return _hashCode;
	}

	/** 
	 * Returns whether this category equals the given object. 
	 * Equivalence up to variable names is handled  
	 * as long as features and predicates are in the same order.
	 * The implementation calls equalsNoLF(obj, varMap, varMap2).
	 */
	public boolean equals(Object obj) { return equals(obj, true); }

	// checks equality, with a flag whether to check the LFs
	private boolean equals(Object obj, boolean checkLF) {
		// NB: The following line can be uncommented, in order to 
		//     turn off category equality checking; it hasn't been 
		//     put on a preferences switch, due to efficiency concerns.
		// if (true) return super.equals(obj);
		if (obj == this) return true;
		if (obj == null) return false;
		if (obj.getClass() != this.getClass()) { return false; }
		return true;
	}
}
