///////////////////////////////////////////////////////////////////////////////
//Copyright (C) 2003-5 Jason Baldridge and University of Edinburgh (Michael White)

//This library is free software; you can redistribute it and/or
//modify it under the terms of the GNU Lesser General Public
//License as published by the Free Software Foundation; either
//version 2.1 of the License, or (at your option) any later version.

//This library is distributed in the hope that it will be useful,
//but WITHOUT ANY WARRANTY; without even the implied warranty of
//MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//GNU Lesser General Public License for more details.

//You should have received a copy of the GNU Lesser General Public
//License along with this program; if not, write to the Free Software
//Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.


package edinburgh.ccg.ccgCheck.utils;

import edinburgh.ccg.ccgCheck.utils.unify.GUnifier;
import edinburgh.ccg.ccgCheck.utils.unify.Substitution;
import edinburgh.ccg.ccgCheck.utils.unify.UnifyFailure;

/**
 * A non-recursive representation of complex categories.
 * 
 * @author Jason Baldridge
 * @author Michael White
 * @version $Revision: 1.12 $, $Date: 2006/10/12 21:20:48 $
 */
public final class ComplexCat extends AbstractCat {

	private static final long serialVersionUID = 1L;

	private TargetCat _target;

	private ArgStack _args;

	/** Constructor with target and single arg. */
	public ComplexCat(TargetCat target, Arg arg) {
		this(target, new ArgStack(arg));
	}

	/** Constructor with target, arg stack and LF. */
	public ComplexCat(TargetCat target, ArgStack args) {
		_target = target;
		if (args.size() < 1) {
			System.out.println("WARNING!!! Creating a ComplexCat with"
					+ " empty argument stack!");
		}
		_args = args;
	}

	/**
	 * Returns the target category of this category.
	 */
	public TargetCat getTarget() {
		return _target;
	}

	public Arg getArg(int pos) {
		return _args.get(pos);
	}

	public Arg getOuterArg() {
		return _args.getLast();
	}

	public Category getResult() {
		return getSubResult(arity() - 1);
	}

	public Category getSubResult(int upto) {
		if (upto == 0) {
			return _target;
		} else {
			return new ComplexCat(_target, _args.subList(0, upto));
		}
	}

	public ArgStack getArgStack() {
		return _args;
	}

	public ArgStack getArgStack(int from) {
		return _args.subList(from);
	}

	public boolean containsDollarArg() {
		return _args.containsDollarArg();
	}

	public boolean containsSetArg() {
		return _args.containsSetArg();
	}

	public void add(Arg a) {
		_args.add(a);
	}

	public void add(ArgStack as) {
		_args.add(as);
	}

	public void addBeforeEnd(ArgStack as) {
		int size = _args.size();
		if (size < 1) {
			add(as);
		} else {
			_args.addAt(as, size - 1);
		}
	}

	public void addFront(ArgStack as) {
		_args.addFront(as);
	}

	public void insertFront(ArgStack as) {
		_args.insertFront(as);
	}

	public void insertEnd(ArgStack as) {
		_args.insertEnd(as);
	}

	public void set(int index, Arg c) {
		_args.set(index, c);
	}

	public void setOuterArgument(Arg c) {
		_args.setLast(c);
	}

	public int arity() {
		return _args.size();
	}

	public Category copy() {
		return new ComplexCat((TargetCat) _target.copy(), _args.copy());
	}

	public Category shallowCopy() {
		return new ComplexCat(_target, _args);
	}

	public void forall(CategoryFcn f) {
		f.forall(this);
		_target.forall(f);
		_args.forall(f);
	}

	public void unifyCheck(Object u) throws UnifyFailure {
		if (u instanceof ComplexCat) {
			ComplexCat cc = (ComplexCat) u;
			_target.unifyCheck(cc._target);
		}
	}

	/** NB: The LF does not participate in unification. */
	public Object unify(Object u, Substitution sub) throws UnifyFailure {

		if (u instanceof AtomCat && arity() == 1 & containsDollarArg()) {
			sub.makeSubstitution((Dollar) _args.get(0), new ArgStack());
			return GUnifier.unify(_target, (AtomCat) u, sub);
		} else if (u instanceof ComplexCat) {
			ComplexCat cc = (ComplexCat) u;
			ArgStack $args = _args.unify(cc._args, sub);
			Category $target = GUnifier.unify(_target, cc._target, sub);
			if ($args.size() == 0) {
				return $target;
			} else {
				return new ComplexCat((TargetCat) $target, $args);
			}
		} else {
			throw new UnifyFailure();
		}
	}

	// nb: not yet sure about calling setLF methods
	public Object fill(Substitution s) throws UnifyFailure {
		Category $target = (Category) _target.fill(s);
		ArgStack $args = _args.fill(s);
		if ($args.size() == 0) {
			return $target;
		}
		if ($target instanceof TargetCat) {
			return new ComplexCat((TargetCat) $target, $args);
		} else if ($target instanceof ComplexCat) {
			((ComplexCat) $target).add($args);
			return $target;
		} else {
			throw new UnifyFailure();
		}
	}

	public String toString() {
		StringBuffer sb = new StringBuffer();
		//(Chris) Surround the target and the first argument (if many) with brackets
		if (_args.size()>1){
			for (int i=0; i<_args.size()-1; i++){
				sb.append('(');
			}
			sb.append(_target.toString());
			for (int i=0; i<_args.size()-1; i++){
				sb.append(_args.get(i).toString());
				sb.append(')');				
			}
			sb.append(_args.get(_args.size()-1).toString());
		}
		else {
			sb.append(_target.toString()).append(_args.toString());
		}
		return sb.toString();
	}

	/**
	 * Returns the supertag for the category.
	 */
	public String getSupertag() {
		StringBuffer sb = new StringBuffer();
		sb.append(_target.getSupertag()).append(_args.getSupertag());
		return sb.toString();
	}

	public String toTeX() {
		// Preferences prefs = Preferences.userNodeForPackage(TextCCG.class);
		// boolean showSem = prefs.getBoolean(SHOW_SEMANTICS, false);
		StringBuffer sb = new StringBuffer();
		sb.append(_target.toTeX()).append(_args.toTeX());
		return sb.toString();
	}
}
