//////////////////////////////////////////////////////////////////////////////
//Copyright (C) 2011-12 Christos Christodoulopoulos and Tejaswini Deoskar

//This library is free software; you can redistribute it and/or
//modify it under the terms of the GNU Lesser General Public
//License as published by the Free Software Foundation; either
//version 3 of the License, or (at your option) any later version.

//This library is distributed in the hope that it will be useful,
//but WITHOUT ANY WARRANTY; without even the implied warranty of
//MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//GNU Lesser General Public License for more details.

//You should have received a copy of the GNU Lesser General Public
//License along with this program; if not, write to the Free Software
//Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

package edinburgh.ccg.parse;


import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import edinburgh.ccg.probModel.util.BinaryTree;
import edinburgh.ccg.util.Options;


/**
 * Represents each chart cell.
 * Should be used for creating the final parse tree.
 * @author Christos Christodoulopoulos
 *
 */
public class Cell {
	//Viterbi derivation which contains both the probability (score) of the category 
	//and back pointers to the children.
	protected HashMap<String, Derivation> vitDerivation;

	protected String word;

	private int coordX, coordY;
	private String[] punctSigns = {".","!",",",";",":","LRB","RRB","LCB","RCB" };
	private List<String> punctList = new ArrayList<String>();


	private Options o = Options.getOptions();

	public Cell(int x, int y){
		vitDerivation = new HashMap<String, Derivation>();
		coordX = x;
		coordY = y;
		for (String sign:punctSigns)
			punctList.add(sign);
	}

	public int size(){
		return vitDerivation.size();
	}
	
	public double getViterbiScore(String cat){
		if(vitDerivation.containsKey(cat)){
			return vitDerivation.get(cat).score;
		}
		else { return 0;}
	}

	/**
	 * Insert leaf nodes for Viterbi (Essentially the same as insertInsideLeaf)
	 */
	public void insertViterbiLeaf(String word, String tag, String result, double derivScore, int index) {
		double newscore = derivScore;
		String resStr = result.toString();
		this.word = word;
		TerminalDerivation deriv = new TerminalDerivation(word, tag, result, derivScore, index);

		if (vitDerivation.containsKey(resStr)){
			double score = vitDerivation.get(resStr).score;
			if(newscore > score){
				vitDerivation.put(resStr, deriv);
			}
		}
		else {
			vitDerivation.put(resStr, deriv);
		}
	}

	/**
	 * Binary Rules (for Viterbi parsing)
	 * cat1 is always the left cell and cat2 is always the right cell being combined. 
	 */
	public void insertVit(	String result, BinaryTree leftTree, 
							Cell cat1Cell, BinaryTree rightTree, 
							Cell cat2Cell, String exp, 
							double derivScore, double vit1, double vit2, 
							String resultKey) {
		//Calculate new derivation score, and make a derivation. 
		double vitScore = derivScore * vit1 *vit2;
		String derivKey;
		if (exp.equals(BinaryTree.leftHeadExp))
			derivKey = resultKey+"|*"+leftTree.cat()+"+"+rightTree.cat();
		else derivKey = resultKey+"|"+leftTree.cat()+"+*"+rightTree.cat();

		BinaryDerivation deriv = new BinaryDerivation(result, leftTree, cat1Cell, 
				rightTree, cat2Cell, exp, vitScore, derivKey);

		if (vitDerivation.containsKey(result)){
			double score = vitDerivation.get(result).score;
			if(vitScore > score) vitDerivation.put(result, deriv);
		}
		else vitDerivation.put(result, deriv);
	}

	/**
	 * For Unary Rules (Viterbi parsing)
	 */
	public void insertVit(  String result, BinaryTree headTree, Cell headCell,
							double derivScore, double vit1, String resultKey) {
		String derivKey = resultKey+"|"+headTree.cat();		
		double vitScore = derivScore * vit1;
		UnaryDerivation deriv = new UnaryDerivation(result, headTree, headCell, vitScore, derivKey);

		if (vitDerivation.containsKey(result)){
			double score = vitDerivation.get(result).score;
			if(vitScore > score) vitDerivation.put(result, deriv);
		}
		else vitDerivation.put(result, deriv);
	}

	/**
	 * Do a top down pass and extract Viterbi derivation.
	 */
	public Derivation getVitDeriv(String res){
		HashMap <String, Double> tempHash = new HashMap<String, Double>();

		for (Map.Entry<String, Derivation> entry : vitDerivation.entrySet() ) {
			tempHash.put(entry.getKey(), entry.getValue().score);
		}

		//Get the entry with the max score
		Map.Entry<String, Double> maxEntry = null;
		for (Map.Entry<String, Double> entry : tempHash.entrySet() ) {
			if (!res.equals("TOP")) {
				//Check if the entry is contained in the head category (entry=sister)
				//or if it contains the result category (entry=head)
				if (!entry.getKey().equals(res)) continue;
			}
			if (maxEntry == null || entry.getValue().compareTo(maxEntry.getValue()) > 0) { 
				maxEntry = entry; 
			}
		}
		return vitDerivation.get(maxEntry.getKey());
	}

	/**
	 * The pruning algorithm. <br>
	 * We set the pruning value relative to the list's size (Roark 2001):<br>
	 * "we want the beam to be very wide id there are few analyses...but
	 * relatively narrow if many analyses have been advanced". <br>
	 * The parse will be discarded if P < P_best*f(alpha, results.size()).
	 */
	public void prune(){
		double bestScore = 0;
		for (Map.Entry<String, Derivation> entry : vitDerivation.entrySet() ) {
			if (bestScore < entry.getValue().score){ 
				bestScore = entry.getValue().score ;				
			}
		}

		double f = o.getAlpha()* vitDerivation.entrySet().size(); //*Math.pow((double)results.size(), 2);
		double pruneVal = bestScore*f;

		for (Map.Entry<String, Derivation> entry : vitDerivation.entrySet() ) {
			if (entry.getValue().score < pruneVal ){
				vitDerivation.remove(entry);
			}
		}
	}
	
	public int getCoordX(){
		return coordX;
	}

	public String toString(){
		String str = "["+coordX+","+coordY+"]:";
		str += vitDerivation.keySet();
		return str;
	}
}
