//////////////////////////////////////////////////////////////////////////////
//Copyright (C) 2003-7 Jason Baldridge, Gann Bierner and Michael White
//Copyright (C) 2011-12 Christos Christodoulopoulos and Tejaswini Deoskar

//This library is free software; you can redistribute it and/or
//modify it under the terms of the GNU Lesser General Public
//License as published by the Free Software Foundation; either
//version 2.1 of the License, or (at your option) any later version.

//This library is distributed in the hope that it will be useful,
//but WITHOUT ANY WARRANTY; without even the implied warranty of
//MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//GNU Lesser General Public License for more details.

//You should have received a copy of the GNU Lesser General Public
//License along with this program; if not, write to the Free Software
//Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

package edinburgh.ccg.parse;

import java.util.ArrayList;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import edinburgh.ccg.ccgCheck.CategoryCreator;
import edinburgh.ccg.ccgCheck.grammar.BackwardApplication;
import edinburgh.ccg.ccgCheck.grammar.BackwardComposition;
import edinburgh.ccg.ccgCheck.grammar.BackwardSubstitution;
import edinburgh.ccg.ccgCheck.grammar.ForwardApplication;
import edinburgh.ccg.ccgCheck.grammar.ForwardComposition;
import edinburgh.ccg.ccgCheck.grammar.ForwardSubstitution;
import edinburgh.ccg.ccgCheck.grammar.RuleGroup;
import edinburgh.ccg.ccgCheck.utils.Category;
import edinburgh.ccg.probModel.ProbModel;
import edinburgh.ccg.probModel.util.BinaryNode;
import edinburgh.ccg.probModel.util.BinaryTree;
import edinburgh.ccg.util.Options;

/**
 * An implementation of the table (or chart) used for CKY.
 * No check for CCG-legal unifications; only derivations seen in the training data are allowed.
 * 
 * @author Christos Christodoulopoulos
 */
public class Chart {

	/** The chart. */
	protected Cell[][] _table;

	/** Its size. */
	protected int _size;
	
	private ProbModel _probModel;
	
	private Options o = Options.getOptions();
	
	private Map<String, Map<String, Map<String, List<String>>>> allowedBinaryDerivations;
	private Map<String, List<String>> allowedUnaryDerivations;
	
	//Used for CCG-check
	private CategoryCreator c;
	private RuleGroup ruleGroup;

	/** Constructor. 
	 * @param binaryRules 
	 * @param unaryRules */
	public Chart(int s, ProbModel p) {
		_size = s;
		_table = new Cell[_size][_size];
		_probModel = p;
		allowedBinaryDerivations = p.allowedBinaryDerivations;
		allowedUnaryDerivations = p.allowedUnaryDerivations;
		if (o.checkCCG()) {
			c = new CategoryCreator();
			ruleGroup = new RuleGroup();
			ruleGroup.addRule(new ForwardApplication());
			ruleGroup.addRule(new BackwardApplication());
			ruleGroup.addRule(new ForwardComposition());
			ruleGroup.addRule(new BackwardComposition());
			ruleGroup.addRule(new ForwardSubstitution());
			ruleGroup.addRule(new BackwardSubstitution());
		}
	}
	
	
	public void initChart(){
		for (int j=0; j<_size; j++) {
			for (int i=0; i<_size; i++) {
			_table[j][i] = new Cell(j, i);
			}
		}
	}

	/** Returns the edge map for a given cell (ensuring non-null). */
	protected Cell get(int x, int y) {
		if (_table[x][y] == null) _table[x][y] = new Cell(x,y);
		return _table[x][y];
	}
	
	/** Inserts edges into (x,y) that result from applying unary rules to those already in (x,y). */ 
	protected void insertCell(int x, int y) {
		if (_table[x][y] == null) return;

		Map<String, Derivation> derivsMap = _table[x][y].vitDerivation;
		Set <String> inputs = new HashSet <String> (derivsMap.keySet());
		for (String input : inputs) {
			BinaryTree inputTree = derivsMap.get(input).resultTree();
			//No point searching any further
			if (!allowedUnaryDerivations.containsKey(input)) continue;
			List<String> resultsFromProbModel = allowedUnaryDerivations.get(input);
			for (String result:resultsFromProbModel){
				String derivationKey = _probModel.getIOKey(result, inputTree);
				double vit1 = _table[x][y].getViterbiScore(input);
				double derivScore = _probModel.scoreDerivation(result,inputTree);
				if(derivScore == 0.0) continue;
				_table[x][y].insertVit(result, inputTree, _table[x][y], derivScore, vit1, derivationKey);
			}
		}
	}

	/** Inserts edges into (x3,y3) resulting from combining those in (x1,y1) and (x2,y2). */
	protected void insertCell(int x1, int y1, int x2, int y2, int x3, int y3) {
		if (_table[x1][y1] == null) return;
		if (_table[x2][y2] == null) return;

		for (String cat1 : _table[x1][y1].vitDerivation.keySet()) {
			BinaryTree leftTree = _table[x1][y1].vitDerivation.get(cat1).resultTree();
			//No point searching any further		
			if (!allowedBinaryDerivations.containsKey(cat1)) continue;
			for (String cat2 : _table[x2][y2].vitDerivation.keySet()) {
				BinaryTree rightTree = _table[x2][y2].vitDerivation.get(cat2).resultTree();
				Map<String, Map<String, List<String>>> rightResMap = allowedBinaryDerivations.get(cat1);
				//No point searching any further
				if (!rightResMap.containsKey(cat2)) continue;
				//**** CCG UNIFICATION ****
				List<String> resultsFromCCG = null;
				if (o.checkCCG()) {
					//Make sure CategoryCreator and RuleGroup have been created before
					Category ccgCat1 = c.createCat(cat1);
					Category ccgCat2 = c.createCat(cat2);
					List<Category> resCats = ruleGroup.applyBinaryRules(ccgCat1, ccgCat2);
					//If no results, we can skip the rest
					if (resCats.isEmpty()) continue;
					resultsFromCCG = new ArrayList<String>(resCats.size());
					for (Category r:resCats){
						resultsFromCCG.add(r.toString());
					}
				}
				Map<String, List<String>> resExpMap = rightResMap.get(cat2);
				for (String expansionFromModel:resExpMap.keySet()){
					List<String> resultsFromProbModel = resExpMap.get(expansionFromModel);
					for (String result:resultsFromProbModel){
						//Check for CCG-legal results
						if (o.checkCCG()) if (!resultsFromCCG.contains(result)) continue;
						double derivScore = _probModel.scoreDerivation(result, leftTree, rightTree, expansionFromModel);
						double vit1 = _table[x1][y1].getViterbiScore(cat1.toString());
						double vit2 = _table[x2][y2].getViterbiScore(cat2.toString());
						String derivationKey = _probModel.getIOKey(result, leftTree, rightTree, expansionFromModel);
						if(derivScore == 0.0) continue;
						_table[x3][y3].insertVit(result, leftTree, _table[x1][y1], 
								rightTree, _table[x2][y2], expansionFromModel, 
								derivScore, vit1, vit2, derivationKey);
					}
				}
			}
		}
	}
	
	public BinaryTree getViterbiParse() throws ParseException{
		if (_table[0][_size-1].vitDerivation.isEmpty()) throw new ParseException("No vit. parse");
		//Top tree node
		Derivation deriv = _table[0][_size-1].getVitDeriv("TOP");
		BinaryTree tree = deriv.resultTree();
		tree.parent = new BinaryNode();
		return tree;
	}
	
	/** Returns the number of entries in each cell in the chart. */
	public String toString() {
		StringBuffer sb = new StringBuffer();
		for (int i = 0; i < _size; i++) {
			for (int j = 0; j < _size; j++) {
				sb.append(get(i, j).size()).append('\t');
			}
			sb.append('\n');
		}
		return sb.toString();
	}
}
