///////////////////////////////////////////////////////////////////////////////
//Copyright (C) 2003-7 Jason Baldridge, Gann Bierner and Michael White
//Copyright (C) 2011-12 Christos Christodoulopoulos and Tejaswini Deoskar

//This library is free software; you can redistribute it and/or
//modify it under the terms of the GNU Lesser General Public
//License as published by the Free Software Foundation; either
//version 2.1 of the License, or (at your option) any later version.

//This library is distributed in the hope that it will be useful,
//but WITHOUT ANY WARRANTY; without even the implied warranty of
//MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//GNU Lesser General Public License for more details.

//You should have received a copy of the GNU Lesser General Public
//License along with this program; if not, write to the Free Software
//Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

package edinburgh.ccg.parse;

import java.io.BufferedWriter;
import java.io.IOException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.StringTokenizer;

import edinburgh.ccg.probModel.ProbModel;
import edinburgh.ccg.probModel.ProbModelBase;
import edinburgh.ccg.probModel.ProbModelLexCat;
import edinburgh.ccg.probModel.util.BinaryTree;
import edinburgh.ccg.probModel.util.PargStructure;
import edinburgh.ccg.util.Logger;
import edinburgh.ccg.util.Options;
import edinburgh.ccg.util.StringUtils;
import edu.stanford.nlp.tagger.maxent.MaxentTagger;

/**
 * The parser is a CKY chart parser for CCG.
 *
 * @author      Jason Baldridge
 * @author      Gann Bierner
 * @author      Michael White
 * @author 		Christos Christodoulopoulos
 */
public class Parser {
	/**(Chris) The statistical model (after Hockenmaier 2003). */
	protected final ProbModel probModel;

	/**(Chris) The tagged sentence. */
	protected String taggedSentence;

	/** The chart. */
	public static Chart theTable;

	private String posTagger = "left3words-wsj-0-18.tagger";

	private Options o = Options.getOptions();

	private StringUtils sUtil = new StringUtils();
	
	private Logger log;

	public static HashMap<String, Double> expectation = new HashMap <String, Double>();

	/** Constructor. */
	public Parser(Logger log) {
		if(o.getModelType().equals("base")){
			probModel = new ProbModelBase(log);
		}
		else if (o.getModelType().equals("lexcat")){
			probModel = new ProbModelLexCat(log);
		}
		else { //Default model type is base
			probModel = new ProbModelBase(log);
		}
		this.log = log;
	}
	
	/**
	 * Parses a string.<br>
	 * Switches between EM re-estimation (-em option) and 1-best
	 * Viterbi parsing.
	 *
	 * @param s the string
	 * @exception ParseException thrown if a parse can't be found for the
	 *            entire string
	 */
	public void parse(String s) throws ParseException {
		try {
			StringTokenizer st = new StringTokenizer(s);
			List<String> words = new ArrayList<String>();
			List<String> tags = new ArrayList<String>();
			ArrayList<Map<String,Double>> results;
			while (st.hasMoreTokens()) {words.add(st.nextToken());}
			if (o.useTags()) {
				MaxentTagger.init(o.getModelsDir()+"/"+posTagger);
				taggedSentence = MaxentTagger.tagString(s);
				st = new StringTokenizer(taggedSentence);
				while (st.hasMoreTokens()) {tags.add(sUtil.extractTag(st.nextToken()));}
				results = probModel.getCategories(words,tags);
			}
			else results = probModel.getCategories(words);
			theTable = new Chart(words.size(), probModel);
			theTable.initChart();
			for (int i=0; i<results.size(); i++){
				Map<String,Double> catMap = results.get(i);
				for (String cat:catMap.keySet()){
					String leaf = cat;
					if (o.useTags()) theTable._table[i][i].insertViterbiLeaf(words.get(i), tags.get(i), leaf, catMap.get(cat), i);
					else theTable._table[i][i].insertViterbiLeaf(words.get(i), null, leaf, catMap.get(cat), i);
				}				
			}
			// do parsing
			parse(theTable, words.size());
		} 
		catch (LexException e) {
			throw new ParseException("Unable to retrieve lexical entries:\n\t"+e.toString());
		} 
		catch (Exception e) {e.printStackTrace();}
	}

	/**
	 * CKY parsing: Used for Inside-Outside re-estimation
	 * and 1-best Viterbi parsing
	 */
	private void parse(Chart table, int size) throws ParseException {
		boolean to = true;
		if (o.getTimeoutMins() == 0) to = false;
		boolean print = false;
		//Convert timeout to miliseconds
		long endTimeMillis = (long) (System.currentTimeMillis() + o.getTimeoutMins()*60000);
		double maxLoops = 0;
		int numLoops = 0;	

		for (int j=1; j<size; j++) {
			for (int i=j-1; i>=0; i--) {
				for (int k=i; k<j; k++) {
					maxLoops++;
				}
			}
		}
		
		log.print("- Parsing...");
		//############# 1-best Viterbi parsing ####################
		// Start with unary rules on the diagonal
		for (int i=0; i<size; i++) {
			table.insertCell(i,i);
		}
		for (int j=1; j<size; j++) {
			for (int i=j-1; i>=0; i--) {
				for (int k=i; k<j; k++) {
					//Progress monitor
					numLoops++;
					double progress = (numLoops*100)/maxLoops;
					if (progress % 10 > 0 && progress % 10 < 1) {
						if (print) log.print(".");
						print = false;
					}
					else if (progress % 10 > 9) print = true;
					//END progress monitor
					
					//Binary					
					table.insertCell(i,k, k+1,j, i,j);
					if (to && System.currentTimeMillis() > endTimeMillis) 
						throw new ParseException("Parse timed out");
				}
				//Unary
				table.insertCell(i,j);
				if (to && System.currentTimeMillis() > endTimeMillis) 
					throw new ParseException("Parse timed out");
			}
		}

		topNodeRerank(table.get(0, size-1));
	}
	
	//For TOP Node
	private void topNodeRerank(Cell topCell) {
		for (String res:topCell.vitDerivation.keySet()){
			Derivation deriv = topCell.vitDerivation.get(res);
			double topExpProb = probModel.getHeadCatProb(BinaryTree.unaryExp,deriv.resultTree,"TOP" );
			topCell.vitDerivation.get(res).score *= topExpProb;
		}
	}
	
	/**(Chris) return the tagged sentence. */
	public String getTaggedSentence(){
		return taggedSentence;
	}

	public void printViterbiParse(int sentNum, BufferedWriter parsesOut, 
			BufferedWriter depsOut) throws ParseException{
		BinaryTree tree = theTable.getViterbiParse();
		int size = theTable._size;
		PargStructure[] headList = new PargStructure[size]; 
		headList = tree.toHeadDeps(headList);
		try {
			parsesOut.write("Sentence "+sentNum+":\n");
			parsesOut.write(tree.toString());
			parsesOut.newLine();
			parsesOut.flush();
			depsOut.write("<s id=\""+sentNum+"\"> "+(size-1));
			depsOut.newLine();
			for (int sisIndex = 0; sisIndex < size; sisIndex++){
				if (headList[sisIndex] == null) continue;
				for (int head = 0; head < headList[sisIndex].size(); head++){
					depsOut.write(sisIndex+"\t"+headList[sisIndex].toString(head));
					depsOut.newLine();
				}
			}
			depsOut.write("<\\s>");
			depsOut.newLine();
		}
		catch(IOException e) {
			e.printStackTrace();
		}

	}
}

