//////////////////////////////////////////////////////////////////////////////
//Copyright (C) 2011-12 Christos Christodoulopoulos and Tejaswini Deoskar

//This library is free software; you can redistribute it and/or
//modify it under the terms of the GNU Lesser General Public
//License as published by the Free Software Foundation; either
//version 3 of the License, or (at your option) any later version.

//This library is distributed in the hope that it will be useful,
//but WITHOUT ANY WARRANTY; without even the implied warranty of
//MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//GNU Lesser General Public License for more details.

//You should have received a copy of the GNU Lesser General Public
//License along with this program; if not, write to the Free Software
//Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

package edinburgh.ccg.probModel;

import java.io.BufferedReader;
import java.io.FileReader;
import java.io.IOException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.StringTokenizer;

import edinburgh.ccg.parse.LexException;
import edinburgh.ccg.probModel.util.BinaryTree;
import edinburgh.ccg.util.Options;


public abstract class ProbModel {
	protected HashMap<String,Double> wordProb;
	protected HashMap<String,Double> expProb;
	protected HashMap<String,Double> headCatProb;
	protected HashMap<String,Double> sisterCatProb;
	protected HashMap<String,Double> sisterLexCatProb; //For model lexcat and above
	protected HashMap<String,Double> sisterHeadWordProb; //For head word models 
	
	protected HashMap<String, List<String>> wordCats;

	protected String modelsDir;
	protected String wordProbModelFileName;
	protected String expProbModelFileName;
	protected String headCatProbModelFileName;
	protected String sisterCatProbModelFileName;
	protected String sisterLexCatProbModelFileName; //For lexcat model and above
	protected String sisterHeadWordProbModelFileName; //For Head Word models
	
	public Map<String, Map<String, Map<String, List<String>>>> allowedBinaryDerivations;
	public Map<String, List<String>> allowedUnaryDerivations;

	private Options o = Options.getOptions();

	/*
	 * Read in the word Prob model file from the models directory
	 * This is the same for all model types, and hence is in the abstract class ProbModel
	 */
	protected void readWordProbModel(String dir, String fileName){
		BufferedReader in;
		String line = null;
		StringTokenizer tok;
		String key;
		double value;
		try {
			in = new BufferedReader(new FileReader(dir+"/"+fileName));
			// Read first line
			line = in.readLine();
			while (line!=null && line!=""){
				tok = new StringTokenizer(line);
				key=tok.nextToken();
				value=Double.parseDouble(tok.nextToken());
				wordProb.put(key, value);
				String word = key.split("\\|")[0];
				String cat = key.split("\\|")[1];
				if (wordCats.containsKey(word)){
					List<String> cats = wordCats.get(word);
					if (!cats.contains(cat)) cats.add(cat);
				}
				else {
					List<String> cats = new ArrayList<String>(1);
					cats.add(cat);
					wordCats.put(word, cats);
				}
				// Read the next line
				line = in.readLine();
			}
		}
		catch (IOException e){e.printStackTrace();}
	}


	/*
	 * Read in a model file into the HashMap probMap.
	 * The format or the model type does not matter at this point 
	 */

	protected void readModel(String dir, String fileName, HashMap<String,Double> probMap){
		BufferedReader in;
		String line = null;
		StringTokenizer tok;
		String key;
		double value;
		try {
			in = new BufferedReader(new FileReader(dir+"/"+fileName));
			// Read first line
			line = in.readLine();
			while (line!=null && line!=""){
				tok = new StringTokenizer(line);
				key=tok.nextToken();
				value=Double.parseDouble(tok.nextToken());
				// Populate the probMap
				probMap.put(key, value);
				// Read the next line
				line = in.readLine();
			}
		}
		catch (IOException e){e.printStackTrace();}
	}
	
	protected void createAllowedDerivations(){
		BufferedReader in;
		String line = null;
		try {
			//First open headCatProbModel to get the unary derivations
			in = new BufferedReader(new FileReader(modelsDir+"/"+headCatProbModelFileName));
			while ((line=in.readLine())!=null){
				line = line.substring(0,line.indexOf(' '));
				//Order is head|exp|result
				String[] splits = line.split("\\|");
				//We need just the unary derivations now
				if (!splits[1].equals(BinaryTree.unaryExp)) continue;
				//Filter out TOP derivations
				if (splits[2].equals("TOP")) continue;
				String res = splits[2];
				String inCat = splits[0];
				List<String> results;
				if (allowedUnaryDerivations.containsKey(inCat)){
					results = allowedUnaryDerivations.get(inCat);
					if (!results.contains(res)) results.add(res);
				}
				else {
					results = new ArrayList<String>(1);
					results.add(res);
					
				}
				allowedUnaryDerivations.put(inCat, results);
			}
			in.close();
			//Now open sisterCatProbModel to get the binary derivations
			in = new BufferedReader(new FileReader(modelsDir+"/"+sisterCatProbModelFileName));
			while ((line=in.readLine())!=null){
				line = line.substring(0,line.indexOf(' '));
				//Order is sister|head|exp|result
				String[] splits = line.split("\\|");
				String exp = splits[2];
				if (exp.equals(BinaryTree.unaryExp)) continue;
				String sis = splits[0];
				String head = splits[1];
				String res = splits[3];
				String inLeft, inRight;
				if (exp.equals(BinaryTree.leftHeadExp)){
					inLeft = head; 
					inRight = sis;
				}
				else {
					inLeft = sis;
					inRight = head;
				}

				if (allowedBinaryDerivations.containsKey(inLeft)){
					//Check if this category has ever been seen with the right category
					if (allowedBinaryDerivations.get(inLeft).containsKey(inRight)){
						//If so add the result to the list
						List<String> results;
						if (allowedBinaryDerivations.get(inLeft).get(inRight).containsKey(exp)){
							results = allowedBinaryDerivations.get(inLeft).get(inRight).get(exp);
							if (!results.contains(res)) results.add(res);
						}
						else {
							results = new ArrayList<String>(1);
							results.add(res);
							
						}
						allowedBinaryDerivations.get(inLeft).get(inRight).put(exp, results);
					}
					else {
						List<String> results = new ArrayList<String>(1);
						results.add(res);
						Map<String, List<String>> tmpMap = new HashMap<String, List<String>>();
						tmpMap.put(exp, results);
						allowedBinaryDerivations.get(inLeft).put(inRight, tmpMap);
					}
				}
				else {
					List<String> results = new ArrayList<String>(1);
					results.add(res);
					Map<String, List<String>> tmpMap = new HashMap<String, List<String>>();
					tmpMap.put(exp, results);
					Map<String, Map<String, List<String>>> tmpMap2 = new HashMap<String, Map<String, List<String>>>();
					tmpMap2.put(inRight, tmpMap);
					allowedBinaryDerivations.put(inLeft, tmpMap2);
				}
			}
		}
		catch (IOException e){e.printStackTrace();}
	}

	/**
	 * Get the categories (and lexical probabilities) for each word in the lexicon.
	 * Now returns a list of maps from cats to probabilities for each word.
	 */
	public ArrayList<Map<String,Double>> getCategories(List<String> words, List<String> tags) throws LexException{
		ArrayList<Map<String,Double>> allWordCats = new ArrayList<Map<String,Double>>();
		for (int i = 0; i < words.size(); i++){
			String word = words.get(i);
			String tag = tags.get(i);
			Map<String,Double> curWordCats = new HashMap<String,Double>();
			if (wordCats.containsKey(word)){
				List<String> cats = wordCats.get(word);
				for (String cat:cats) curWordCats.put(cat, wordProb.get(word+"|"+cat));
			}
			//Back-off to PoS tags
			else if (o.useTags() && wordCats.containsKey(tag)){
				List<String> cats = wordCats.get(tag);
				for (String cat:cats) curWordCats.put(cat, wordProb.get(tag+"|"+cat));
			}
			else throw new LexException ("Unknown word: "+word);

			allWordCats.add(curWordCats);
		}
		return allWordCats;
	}

	public ArrayList<Map<String,Double>> getCategories(List<String> words) throws LexException{
		ArrayList<Map<String,Double>> allWordCats = new ArrayList<Map<String,Double>>();
		for (int i = 0; i < words.size(); i++){
			String word = words.get(i);
			Map<String,Double> curWordCats = new HashMap<String,Double>();
			if (wordCats.containsKey(word)){
				List<String> cats = wordCats.get(word);
				for (String cat:cats) curWordCats.put(cat, wordProb.get(word+"|"+cat));
			}
			else throw new LexException ("Unknown word: "+word);

			allWordCats.add(curWordCats);
		}
		return allWordCats;
	}

	public double getWordProb(String key) throws LexException{
		double prob = 0;
		if (wordProb.containsKey(key)) return wordProb.get(key);
		if (prob==0 && o.isBackoff()) prob=o.getBackoffProb();
		return prob;
	}

	
	public abstract double getHeadCatProb(String exp, BinaryTree headTree, String resultCat);
	public abstract double scoreDerivation(String result , BinaryTree tree1 , BinaryTree tree2, String exp);
	public abstract double scoreDerivation(String result, BinaryTree tree);
	
	public abstract String getIOKey(String result , BinaryTree tree1 , BinaryTree tree2, String exp);
	public abstract String getIOKey(String result, BinaryTree tree);
	public abstract String getIOKeyUnary(String result, BinaryTree tree);
	public abstract String getIOKeyLeaf(String result, String word); //For lexical nodes

	public abstract String getDerivKey(String result , BinaryTree tree1 , BinaryTree tree2, String exp);
	public abstract String getDerivKey(String result, BinaryTree tree);	
}
