//////////////////////////////////////////////////////////////////////////////
//Copyright (C) 2011-12 Christos Christodoulopoulos and Tejaswini Deoskar

//This library is free software; you can redistribute it and/or
//modify it under the terms of the GNU Lesser General Public
//License as published by the Free Software Foundation; either
//version 3 of the License, or (at your option) any later version.

//This library is distributed in the hope that it will be useful,
//but WITHOUT ANY WARRANTY; without even the implied warranty of
//MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//GNU Lesser General Public License for more details.

//You should have received a copy of the GNU Lesser General Public
//License along with this program; if not, write to the Free Software
//Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

package edinburgh.ccg.probModel;

import java.io.File;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import edinburgh.ccg.probModel.util.BinaryTree;
import edinburgh.ccg.util.Logger;
import edinburgh.ccg.util.Options;


public class ProbModelBase extends ProbModel{

	protected Options o = Options.getOptions();

	// Constructor
	public ProbModelBase(Logger log){

		modelsDir = o.getModelsDir();
		wordProbModelFileName      = o.getWordProbModelFile();
		expProbModelFileName       = o.getExpProbModelFile();
		headCatProbModelFileName   = o.getHeadCatProbModelFile();
		sisterCatProbModelFileName = o.getSisterCatProbModelFile();

		File wordProbModelFile = new File(modelsDir+"/"+wordProbModelFileName);
		File expProbModelFile = new File(modelsDir+"/"+expProbModelFileName);
		File headCatProbModelFile = new File(modelsDir+"/"+headCatProbModelFileName);
		File sisterCatProbModelFile = new File(modelsDir+"/"+sisterCatProbModelFileName);

		wordProb = new HashMap<String,Double>();
		expProb = new HashMap<String,Double>();
		headCatProb = new HashMap<String,Double>();
		sisterCatProb = new HashMap<String,Double>();		
		wordCats = new HashMap<String, List<String>>();
		allowedBinaryDerivations = new HashMap<String, Map<String, Map<String, List<String>>>>();
		allowedUnaryDerivations = new HashMap<String, List<String>>();

		boolean filesExist = wordProbModelFile.exists() && expProbModelFile.exists() 
		&& headCatProbModelFile.exists() && sisterCatProbModelFile.exists();	

		if (o.isTrain() || !filesExist) {
			new ProbModelBaseTrainer(log);
		}
		log.print("Reading statistical model...");
		readWordProbModel(modelsDir,wordProbModelFileName);
		readModel(modelsDir,expProbModelFileName,expProb);
		readModel(modelsDir,headCatProbModelFileName,headCatProb);
		readModel(modelsDir,sisterCatProbModelFileName,sisterCatProb);
		createAllowedDerivations();

		log.println("done");	
	}

	/**
	 * Calculates the binary rule probability: P(A -> B C)
	 */
	public double scoreDerivation(String result, BinaryTree tree1, BinaryTree tree2, String expansion){
		double expProb;
		double headCatProb;
		double sisterCatProb;
		double derivScore;

		BinaryTree headTree, sisterTree;
		if (expansion.equals(BinaryTree.leftHeadExp)){
			headTree = tree1; sisterTree = tree2;
		}
		else {
			headTree = tree2; sisterTree = tree1;
		}
		expProb = getExpProb(expansion, result);
		headCatProb = getHeadCatProb(expansion, headTree, result);
		sisterCatProb = getSisterCatProb(expansion, headTree.cat(), sisterTree.cat(), result);
		derivScore = expProb*headCatProb*sisterCatProb;
		return derivScore;
	}

	/**
	 * Calculates the unary rule probability: P(A -> B)
	 */
	public double scoreDerivation(String result, BinaryTree in){
		double expProb;
		double headCatProb;
		double ruleScore;
		String expansion = BinaryTree.unaryExp;
		expProb = getExpProb(expansion,  result);
		headCatProb = getHeadCatProb(expansion, in, result);
		ruleScore = expProb*headCatProb;
		return ruleScore;
	}

	public double getExpProb(String exp, String cat) {
		// expProb = P(exp|resultCat)
		String key = exp+"|"+cat;
		double prob = 0;
		if (expProb.containsKey(key)) return expProb.get(key);
		if (prob==0 && o.isBackoff()) prob=o.getBackoffProb();
		return prob;
	}

	public double getHeadCatProb(String exp, BinaryTree headTree, String resultCat) {
		// headCatProb = P(H|exp,resultCat)
		String headCat = headTree.cat();
		String key = headCat+"|"+exp+"|"+resultCat;
		double prob = 0;
		if (headCatProb.containsKey(key)) return headCatProb.get(key);
		if (prob==0 && o.isBackoff()) {prob=o.getBackoffProb();System.out.println("Backing off head prob");}
		return prob;
	}

	public double getSisterCatProb(String exp, String headCat, String sisterCat, String resultCat) {
		// sisterCatProb = P(S|H,exp,resultCat)
		String key = sisterCat+"|"+headCat+"|"+exp+"|"+resultCat.toString();
		double prob = 0;
		if (sisterCatProb.containsKey(key)) return sisterCatProb.get(key);
		if (prob==0 && o.isBackoff()) {prob=o.getBackoffProb(); System.out.println("Backing off sister prob");}
		return prob;
	}

	/**
	 * Returns the String with with the inside (and outside) hashes are keyed. 
	 */
	public String getIOKey(String result, BinaryTree tree1, BinaryTree tree2, String expansion){				
		//For the base model, this is all that is required.
		return result ;
	}

	/**
	 * Returns the String with with the inside (and outside) hashes are keyed. 
	 */
	public String getIOKey(String result, BinaryTree in){
		return result;
	}

	public String getIOKeyUnary(String result, BinaryTree in){
		return result;
	}

	/**
	 * For pre-terminal (lexical) categories
	 */
	public String getIOKeyLeaf(String result, String word){
		return result ;
	}

	public String getDerivKey(String result, BinaryTree tree1, BinaryTree tree2, String expansion){
		String cat1Key = tree1.cat();
		String cat2Key = tree2.cat() ;
		if (expansion.equals(BinaryTree.leftHeadExp)){
			return result + "->*" + cat1Key + " + " + cat2Key;
		}
		else{
			return result + "->" + cat1Key + " + *" + cat2Key;
		}
	}

	public String getDerivKey(String result, BinaryTree in){		
		String key = result + " -> " + in.cat();
		return key ;
	}

	public static String removeFeatTypes(String text){
		StringBuffer sb;
		if (text.indexOf('[')>0){
			sb = new StringBuffer(text.substring(0,text.indexOf('[')));
			sb.append(text.substring(text.indexOf(']')+1,text.length()));
			return removeFeatTypes(sb.toString());
		}
		return text;
	}
}
