//////////////////////////////////////////////////////////////////////////////
//Copyright (C) 2011-12 Christos Christodoulopoulos and Tejaswini Deoskar

//This library is free software; you can redistribute it and/or
//modify it under the terms of the GNU Lesser General Public
//License as published by the Free Software Foundation; either
//version 3 of the License, or (at your option) any later version.

//This library is distributed in the hope that it will be useful,
//but WITHOUT ANY WARRANTY; without even the implied warranty of
//MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//GNU Lesser General Public License for more details.

//You should have received a copy of the GNU Lesser General Public
//License along with this program; if not, write to the Free Software
//Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

package edinburgh.ccg.probModel;

import java.io.File;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import edinburgh.ccg.probModel.util.BinaryTree;
import edinburgh.ccg.util.Logger;
import edinburgh.ccg.util.Options;

public class ProbModelLexCat extends ProbModel {

	protected Options o = Options.getOptions();

	// Constructor
	public ProbModelLexCat(Logger log){

		modelsDir = o.getModelsDir();
		wordProbModelFileName      = o.getWordProbModelFile();
		expProbModelFileName       = o.getExpProbModelFile(); 
		headCatProbModelFileName   = o.getHeadCatProbModelFile();
		sisterCatProbModelFileName = o.getSisterCatProbModelFile();
		sisterLexCatProbModelFileName = o.getSisterLexCatProbModelFile();

		File wordProbModelFile = new File(modelsDir+"/"+wordProbModelFileName);
		File expProbModelFile = new File(modelsDir+"/"+expProbModelFileName);
		File headCatProbModelFile = new File(modelsDir+"/"+headCatProbModelFileName);
		File sisterCatProbModelFile = new File(modelsDir+"/"+sisterCatProbModelFileName);
		File sisterLexCatProbModelFile = new File(modelsDir+"/"+sisterLexCatProbModelFileName);

		wordProb = new HashMap<String,Double>();
		expProb = new HashMap<String,Double>();
		headCatProb = new HashMap<String,Double>();
		sisterCatProb = new HashMap<String,Double>();		
		sisterLexCatProb = new HashMap<String,Double>();	
		wordCats = new HashMap<String, List<String>>();
		allowedBinaryDerivations = new HashMap<String, Map<String, Map<String, List<String>>>>();
		allowedUnaryDerivations = new HashMap<String, List<String>>();

		boolean filesExist = wordProbModelFile.exists() && expProbModelFile.exists() 
		&& headCatProbModelFile.exists() && sisterCatProbModelFile.exists() && sisterLexCatProbModelFile.exists();

		//Retrain if the switch is on, or if any (or all) of the files is missing
		if (o.isTrain() || !filesExist) {
			new ProbModelLexCatTrainer(log);
		}	

		log.print("Reading statistical model...");

		readWordProbModel(modelsDir,wordProbModelFileName);
		readModel(modelsDir,expProbModelFileName,expProb);
		readModel(modelsDir,headCatProbModelFileName,headCatProb);
		readModel(modelsDir,sisterCatProbModelFileName,sisterCatProb);
		readModel(modelsDir,sisterLexCatProbModelFileName,sisterLexCatProb);
		createAllowedDerivations();
		log.println("done");
	}

	/**
	 * Calculates the binary rule probability: P(A -> B C). This is different from the base model, due to the
	 * sister cat probabilities being backed off. 
	 */
	public double scoreDerivation(String result, BinaryTree tree1, BinaryTree tree2, String expansion){
		double expProb;
		double headCatProb;
		double sisterCatProb; 
		double sisterLexCatProb; 
		double derivScore;		

		BinaryTree headTree, sisterTree;
		if (expansion.equals(BinaryTree.leftHeadExp)){
			headTree = tree1; sisterTree = tree2;
		}
		else {
			headTree = tree2; sisterTree = tree1;
		}
		expProb = getExpProb(expansion, headTree, result);
		headCatProb = getHeadCatProb(expansion, headTree, result);
		sisterCatProb = getSisterCatProb(expansion, headTree, sisterTree, result);
		sisterLexCatProb = getSisterLexCatProb(expansion, headTree, sisterTree, result);
		derivScore = expProb*headCatProb*sisterCatProb*sisterLexCatProb;
		return derivScore;
	}

	/**
	 * Calculates the unary rule probability: P(A -> B)
	 */
	public double scoreDerivation(String result, BinaryTree in){
		double expProb;
		double headCatProb;
		double ruleScore;
		String expansion = BinaryTree.unaryExp;
		expProb = getExpProb(expansion, in, result);		
		headCatProb = getHeadCatProb(expansion, in, result);
		ruleScore = expProb*headCatProb;
		return ruleScore;
	}

	/**
	 * Backed off the a coarser distribution if the finer category (with lexcat) does not exist.
	 * This is not the standard way that Julia's parser does it. 
	 * She does not backoff in the case of the ExpProb distribution. 
	 */
	public double getExpProbBackedOff(String exp,  BinaryTree inTree, String resultCat) {
		//expProb = P(exp|resultCat, lexcat)
		String lexcat = inTree.lexCat();
		String key = exp+"|"+resultCat;
		String keyLex = exp+"|"+resultCat+"||"+lexcat;
		double prob = 0;
		if (expProb.containsKey(keyLex)) {
			return expProb.get(keyLex);
		}
		else if (expProb.containsKey(key)) {
			return expProb.get(key);
		}
		return prob;
	}


	/**
	 * No backoff to a coarser distribution.
	 * This is the standard way of Julia's parser
	 */
	public double getExpProb(String exp,  BinaryTree inTree, String resultCat) {
		//expProb = P(exp|resultCat, lexcat)
		String lexcat = inTree.lexCat();
		String keyLex = exp+"|"+resultCat+"||"+lexcat;
		double prob = 0;
		if (expProb.containsKey(keyLex)) return expProb.get(keyLex);
		return prob;
	}

	/**
	 * No backing off to a coarser distribution.
	 * This follows Julia's parser
	 */
	public double getHeadCatProb(String exp, BinaryTree headTree, String resultCat) {
		// headCatProb = P(H|exp,resultCat, lexcat)
		String lexcat = headTree.lexCat();
		String headCat = headTree.cat();
		String keyLex = headCat+"|"+exp+"|"+resultCat+"||"+lexcat;
		double prob = 0;
		if (headCatProb.containsKey(keyLex)) return headCatProb.get(keyLex);
		return prob;
	}
	
	/**
	 * This distribution is always backed off by default (as per Julia's LexCat model)
	 */
	public double getSisterCatProb(String exp, BinaryTree headTree, BinaryTree sisterTree, String resultCat) {
		//sisterCatProb = P(S|H,exp,resultCat)
		String lexcat = headTree.lexCat();
		String headCat = headTree.cat();
		String sisterCat = sisterTree.cat();
		String key = sisterCat+"|"+headCat+"|"+exp+"|"+resultCat.toString();
		String keyLex = sisterCat+"|"+headCat+"|"+exp+"|"+resultCat+"||"+lexcat ;
		double prob = 0;
		if (sisterCatProb.containsKey(keyLex)) {
			return sisterCatProb.get(keyLex);
		}
		else {
			if (sisterCatProb.containsKey(key)) {
				return sisterCatProb.get(key);
			}
		}
		return prob;
	}

	public double getSisterLexCatProb(String exp, BinaryTree headTree, BinaryTree sisterTree, String resultCat) {
		//sisterCatProb = P(S|H,exp,resultCat)
		String headCat = headTree.cat();
		String sisterCat = sisterTree.cat();
		String sisterLexCat = sisterTree.lexCat();
		String keyFine = sisterLexCat +"|" + sisterCat+"|"+headCat+"|"+exp+"|"+resultCat;
		String keyCoarse = sisterLexCat;
		double prob = 0;
		if (sisterLexCatProb.containsKey(keyFine)){ 
			return sisterLexCatProb.get(keyFine);
		}
		else {
			if (sisterLexCatProb.containsKey(keyCoarse)){ 
				return sisterLexCatProb.get(keyCoarse);
			}
		}
		return prob;
	}

	/**
	 * Returns the String with with the inside (and outside) hashes are keyed. 
	 */
	public String getIOKey(String result, BinaryTree tree1, BinaryTree tree2, String expansion){
		BinaryTree headTree;
		if (expansion.equals(BinaryTree.leftHeadExp)) 	headTree = tree1; 	
		else headTree = tree2; 	
		String key = result + "||" + headTree.lexCat();
		return key ;
	}
	
	/**
	 * Returns the String with the inside (and outside) hashes are keyed. 
	 */
	public String getIOKey(String result, BinaryTree in){		
		String key = result + "||" + in.lexCat();
		return key ;
	}

	public String getIOKeyUnary(String result, BinaryTree in){		
		String key = result + "||" + in.lexCat();
		return key ;
	}

	/**
	 * For pre-terminal (lexical) categories
	 */
	public String getIOKeyLeaf(String result, String word){
		String key = result  + "||" + result;
		return key;
	}

	public String getDerivKey(String result, BinaryTree tree1, BinaryTree tree2, String expansion){
		String cat1Key = tree1.cat();
		String cat2Key = tree2.cat() ;
		if (expansion.equals(BinaryTree.leftHeadExp)){
			return result + "||" + tree1.lexCat() + "->*" + cat1Key + " + " + cat2Key + "||" + tree2.lexCat();
		}
		else {
			return result + "||" + tree2.lexCat() + "->" + cat1Key+ "||" + tree1.lexCat() + " + *" + cat2Key;
		}
	}

	public String getDerivKey(String result, BinaryTree in){		
		String key = result + "||" + in.lexCat() + "->" + in.cat();
		return key ;
	}
}
