//////////////////////////////////////////////////////////////////////////////
//Copyright (C) 2011-12 Christos Christodoulopoulos and Tejaswini Deoskar

//This library is free software; you can redistribute it and/or
//modify it under the terms of the GNU Lesser General Public
//License as published by the Free Software Foundation; either
//version 3 of the License, or (at your option) any later version.

//This library is distributed in the hope that it will be useful,
//but WITHOUT ANY WARRANTY; without even the implied warranty of
//MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//GNU Lesser General Public License for more details.

//You should have received a copy of the GNU Lesser General Public
//License along with this program; if not, write to the Free Software
//Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

package edinburgh.ccg.probModel;

import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;

import edinburgh.ccg.probModel.util.BinaryLeaf;
import edinburgh.ccg.util.Logger;
import edinburgh.ccg.util.Options;
import edinburgh.ccg.util.Pair;

public abstract class ProbModelTrainer {
	protected File trainFile;
	protected File smoothTrainFile;  

	protected HashMap<String,Double> wordProb;
	protected HashMap<String,Double> expProb;
	protected HashMap<String,Double> headProb;
	protected HashMap<String,Double> sisterProb;
	protected HashMap<String,Double> sisterLexCatProb; //For LexCat model and above
	protected HashMap<String,Double> backOffSisterLexCatProb;
	protected HashMap<String,Double> backOff1SisterProb;

	//Extra for LexCat model
	//Backoff prob maps for Expansion model
	protected HashMap<String,Double> backOffExpProb;
	protected HashMap<String,Double> newExpProb;
	protected HashMap<String,Double> newBackOffExpProb;

	//Backoff prob maps for HeadCat model
	protected HashMap<String,Double> backOffHeadProb;
	protected HashMap<String,Double> newHeadProb;
	protected HashMap<String,Double> newBackOffHeadProb;


	//Backoff prob maps for SisterCat model
	protected HashMap<String,Double> backOffSisterProb;

	protected HashMap<String,Double> newSisterProb; //Interpolated value	
	protected HashMap<String,Double> newBackOffSisterProb; //Interpolated value	
	protected HashMap<String,Double> newSisterLexCatProb; //Interpolated value
	protected HashMap<String,Double> newBackOffSisterLexCatProb;

	// Counters: we only need 6 different counters
	// all the probabilities can be calculated from these
	protected HashMap<String,Integer> wordNum;
	protected HashMap<String,Integer> wordDenom;
	protected HashMap<String,Integer> expDenom;
	protected HashMap<String,Integer> expNum;
	protected HashMap<String,Integer> headNum;
	protected HashMap<String,Integer> sisterNum;

	//Extra counters For Lex Cat model 
	//Backoff counters for Expansion model
	protected HashMap<String,Integer> backOffExpNum;
	protected HashMap<String,Integer> backOffExpDenom;

	//Back off counters for SisterCat model
	protected HashMap<String,Integer> backOffSisterNum;	
	protected HashMap<String,Integer> backOffHeadNum;

	//Counter for Sister LexCat model
	protected HashMap<String,Integer> sisterLexCatNum ;
	protected HashMap<String,Integer> backOffSisterLexCatNum ;
	protected HashMap<String,Integer> backOffSisterLexCatDenom;

	//words->tags->[cats,freq]
	/** Needed to replace rare words with their tags */
	protected Map<String, Map<String, Pair<String, Integer>>> wordTagCatMapCount = new HashMap<String, Map<String, Pair<String, Integer>>>();	

	// Store counts of Words and Tags from smooth training data. 
	public HashMap<String, HashMap <String, Double>> testWordTagFreq;
	public HashMap<String, Integer> testTagFreq;

	protected Options o = Options.getOptions();
	protected Logger log;


	public ProbModelTrainer(Logger log){
		trainFile = new File(o.getTrainFile());
		this.log = log;
		log.println("Creating the statistical model from file: "+trainFile.getName()+"...");

		/* TD: Also read in a POS tagged test file (i.e. the sentences to be parsed. 
		 * This is so that words in the test data can be added to the lexicon, with
		 * appropriate smoothing.  
		 */
		if (o.isSmoothTrain()) smoothTrainFile = new File(o.getSmoothTrainFile());
		trainProbModel(o.isSmoothTrain());
		log.println("done!");		

	}

	protected void trainProbModel(boolean smoothTrain){
		// Initialise the prob maps
		wordProb = new HashMap<String,Double>();
		expProb = new HashMap<String,Double>();
		headProb = new HashMap<String,Double>();
		sisterProb = new HashMap<String,Double>();

		// Initialise the counters
		wordNum = new HashMap<String,Integer>();
		wordDenom = new HashMap<String,Integer>();
		expDenom = new HashMap<String,Integer>();
		expNum = new HashMap<String,Integer>();
		headNum = new HashMap<String,Integer>();
		sisterNum = new HashMap<String,Integer>();		

		try {
			BufferedReader trainingBuffer = new BufferedReader(new FileReader(trainFile));
			BufferedReader testBuffer = null;
			if (smoothTrain) testBuffer = new BufferedReader(new FileReader(smoothTrainFile));

			//TD: training the model includes adding test words and smoothing frequencies 
			// of words in the test corpus.
			trainModel(trainingBuffer, testBuffer); 

		}
		catch (FileNotFoundException e){
			if (smoothTrain) System.err.println("ERROR: Unable to read training or test files"+trainFile + ", " +smoothTrainFile );
			else System.err.println("ERROR: Unable to read training file");
		} 
		catch (IOException e) {e.printStackTrace();}
		clear();
	}


	protected abstract void trainModel(BufferedReader trainBuffer, BufferedReader testBuffer) throws IOException ;

	/**
	 * Lexical counters are the same for all model types	 
	 */
	protected void updateCounts(BinaryLeaf node){
		incrementFreq(wordNum, node.lexNumKey());
		incrementFreq(wordDenom, node.lexCat());
	}


	protected void incrementMapFreq(Map<String, Map<String, Pair<String, Integer>>> map, String word, String tag, String cat){
		Map<String,Pair<String,Integer>> tMap;
		Pair<String,Integer> catFreq;
		if (map.containsKey(word)){
			tMap = map.get(word);
			if (tMap.containsKey(tag))
				catFreq = new Pair<String,Integer>(cat,tMap.get(tag).b+1);
			else catFreq = new Pair<String,Integer>(cat,1);
		}
		else {
			tMap = new HashMap<String,Pair<String,Integer>>();
			catFreq = new Pair<String,Integer>(cat,1);
		}
		tMap.put(tag, catFreq);
		map.put(word,tMap);
	}

	/** Increments the entry $map{$key} by 1, or sets it to 1 if it doesn't exist yet. */  
	protected void incrementFreq(HashMap<String,Integer> map, String key){
		if (key != null && key.length() > 0 && map.containsKey(key)){
			int freq = ((Integer)map.get(key)).intValue();
			map.put(key, new Integer(++freq));
		}
		else  map.put(key, new Integer(1));
	}

	protected void removeRareCats(){
		//Go through catCounter and find all cats with < THRESHOLD counts
		//To avoid concurrency problems just have a list of the things to delete
		List<String> toDelCat = new ArrayList<String>();
		List<String> toDelCatLeaf = new ArrayList<String>();
		List<String> toDelExp = new ArrayList<String>();
		List<String> toDelHead = new ArrayList<String>();
		List<String> toDelSister = new ArrayList<String>();
		List<String> toDelWord = new ArrayList<String>();
		for (String cat:expDenom.keySet()){
			if (expDenom.get(cat) > o.getCatThreshold()) continue;
			toDelCat.add(cat);
			for (String catLeaf:wordDenom.keySet()){
				if (catLeaf.equals(cat)) toDelCatLeaf.add(catLeaf);
			}
			for (String catLeaf:wordNum.keySet()){
				if (catLeaf.endsWith("|"+cat)) toDelWord.add(catLeaf);
			}
			for (String expCat:expNum.keySet()){
				if (expCat.endsWith("|"+cat)) toDelExp.add(expCat);
			}
			for (String headExpCat:headNum.keySet()){
				if (headExpCat.endsWith("|"+cat)) toDelHead.add(headExpCat);
			}
			for (String sisterHeadExpCat:sisterNum.keySet()){
				if (sisterHeadExpCat.endsWith("|"+cat)) toDelSister.add(sisterHeadExpCat);
			}
		}
		//Now go over the lists and delete the entries
		for (String cat:toDelCat) expDenom.remove(cat);
		for (String cat:toDelCatLeaf) wordDenom.remove(cat);
		for (String cat:toDelExp) expNum.remove(cat);
		for (String cat:toDelHead) headNum.remove(cat);
		for (String cat:toDelSister) sisterNum.remove(cat);
	}

	protected void replaceRareWords(){
		List<String> toDelWord = new ArrayList<String>();
		Map<String,Integer> toAddWord = new HashMap<String,Integer>();
		for (String word:wordTagCatMapCount.keySet()){
			Map<String, Pair<String, Integer>> tagMap = wordTagCatMapCount.get(word);
			for (String tag:tagMap.keySet()){
				int freq = tagMap.get(tag).b;
				if (freq>o.getRareWordThreshold()) continue;
				String cat = tagMap.get(tag).a;
				String newKey = tag+"|"+cat;
				if (toAddWord.containsKey(newKey))
					toAddWord.put(newKey, toAddWord.get(newKey)+freq);
				else toAddWord.put(newKey, freq);
				toDelWord.add(word+"|"+cat);
			}
		}
		for (String cat:toDelWord) wordNum.remove(cat);
		wordNum.putAll(toAddWord);
	}

	/**
	 * numMap and denomMap correspond to the numerator and denominator frequencies of probMap
	 * These could be put together into a distribution family like what Julia's parser has.
	 */
	void computeBackOff(HashMap <String, Double> newProbMap, 
			HashMap <String, Double> newBackOffProbMap, 
			HashMap <String, Double> probMap, 
			HashMap <String, Double> backOffProbMap, 
			HashMap <String, Integer> numMap,
			HashMap <String, Integer>denomMap){

		double newProb; //The backed off value
		double prob;
		double backoffProb;
		double lambda;
		String numKey;

		//For each entry in the fine-grained probMap
		for (Map.Entry<String,Double> entry : probMap.entrySet()){
			prob = entry.getValue();
			numKey  = entry.getKey();

			//lambda = getLambda(diversityMap, denomKey, denomFreq);
			lambda = getLambda();

			String backOffKey = getKeyFromMap(numKey, backOffProbMap);

			if(backOffProbMap.containsKey(backOffKey)) {
				backoffProb = backOffProbMap.get(backOffKey);
			}
			else backoffProb = 0;

			newProb = lambda*(prob) + ( 1 - lambda)*backoffProb;
			newProbMap.put(entry.getKey(), newProb);
			newBackOffProbMap.put(backOffKey, ( 1 - lambda)*backoffProb);
		}

	}


	private String getKeyFromMap(String key, Map<String, Double> map){
		String result = null;
		boolean found = false;
		for (String k:map.keySet()){
			if (key. contains(k)) {
				if (!found){
					result = k;
					found = true;
				}
				else {
					//We always need the longest substring of "key"
					if (result.length() < k.length()) result = k;
				}
			}
		}
		return result;
	}


	/**
	 * Calculate the lambda value given a frequency and a diversityMap
	 */
	double getLambda(){
		return 0.01;
	}

	/* Chris:
	 * Can be used for both Double and Integer HashMaps! 
	 */
	protected void computeProbs(HashMap<String,? extends Number> numMap, HashMap<String,Integer> denMap, HashMap<String,Double> probMap){
		// Here the pipe stands for ',' in the joint probabilities 
		int pipeInd;
		String numerator, denominator;
		double probability;

		for (Iterator<String> it = numMap.keySet().iterator(); it.hasNext();){
			numerator = it.next();
			pipeInd = numerator.indexOf('|');
			denominator = numerator.substring(++pipeInd);
			if (denMap.containsKey(denominator)){
				probability = numMap.get(numerator).doubleValue()/(double)denMap.get(denominator);
				probMap.put(numerator, probability);
			}
		}
	}

	/**
	 * Different from computeProbs only in that the things stripped off to get denominator is a double pipe
	 */
	protected void computeLexCatProbs(HashMap<String,? extends Number> numMap, HashMap<String,Integer> denMap, HashMap<String,Double> probMap){
		// Here the pipe stands for ',' in the joint probabilities 
		int pipeInd;
		String numerator, denominator;
		double probability;

		for (Iterator<String> it = numMap.keySet().iterator(); it.hasNext();){
			numerator = it.next();
			pipeInd = numerator.lastIndexOf('|'); //Remove the part of the numKey that is after the '||' indicating lexcat.
			denominator = numerator.substring(0, --pipeInd);
			if (denMap.containsKey(denominator)){
				probability = numMap.get(numerator).doubleValue()/(double)denMap.get(denominator);
				probMap.put(numerator, probability);
			}
		}
	}

	protected void writeModel(String dir, String fileName, HashMap<String,Double> probMap, Boolean append){
		BufferedWriter out;
		try {
			out = new BufferedWriter(new FileWriter (dir+"/"+fileName, append));
			for (Map.Entry<String,Double> entry : probMap.entrySet()){
				out.write(entry.getKey()+" "+entry.getValue().toString());
				out.newLine();
			}
			out.close();
		}
		catch (IOException e){e.printStackTrace();}
	}

	protected void writeHeadMap(String dir, String fileName, Map<String, Map<String, String>> hMap){
		BufferedWriter out;
		try {
			out = new BufferedWriter(new FileWriter (dir+"/"+fileName));
			for (String resCat:hMap.keySet()){
				Map<String, String> tMap = hMap.get(resCat);
				for (String inputs:tMap.keySet()){
					out.write(resCat+"\t"+inputs+"\t"+tMap.get(inputs));
					out.newLine();
				}
			}
			out.close();
		}
		catch (IOException e){e.printStackTrace();}
	}

	protected void writeModels(){
		// Store the new models
		writeModel(o.getModelsDir(),o.getWordProbModelFile(),wordProb, false);
		writeModel(o.getModelsDir(),o.getExpProbModelFile(),expProb, false);
		writeModel(o.getModelsDir(),o.getHeadCatProbModelFile(),headProb, false);
		writeModel(o.getModelsDir(),o.getSisterCatProbModelFile(),sisterProb, false);
	}

	protected void clear(){
		// Clear the HashMaps and collect garbage
		wordProb.clear();
		wordProb = null;
		expProb.clear();
		expProb = null;
		headProb.clear();
		headProb = null;
		sisterProb.clear();
		sisterProb = null;

		System.gc();
	}
}
