package edinburgh.ccg.probModel.util;

import java.util.HashMap;
import java.util.Map;
import java.util.StringTokenizer;

import edinburgh.ccg.util.StringUtils;


/** The class for the binary trees used by the statistical parser.
    (actually, now also flat trees...!!!).

    @author Julia Hockenmaier
    @author Christos Christodoulopoulos
 */
public abstract class BinaryTree implements Cloneable{

	/** The category of this constituent */
	public String cat;

	/** During parsing: indicates the word which is head. */
	public int headpos;

	/** We need to have a parent node in order to be able to read in the trees. */    
	public BinaryTree parent;

	/** The head daughter */
	public BinaryTree head;
	
	/** The non-head daughter */
	public BinaryTree sister;

	/** This indicates the expansion. */
	public String exp;

	/** During training: points to the tree which is built up. */
	public static BinaryTree current;    


	final static String open_brackets = "("; 
	final static String closed_brackets = ")";
	final static String TOP = "TOP";

	/** java.lang.Double.NEGATIVE_INFINITY */
	final static double NEGINF =  java.lang.Double.NEGATIVE_INFINITY;

	public static boolean translateWords = true;
	
	public static final String leftHeadExp = "headLeft";
	public static final String rightHeadExp = "headRight";
	public static final String unaryExp = "unary";
	public static final String leafExp = "leaf";
	
	/** Map from (result) categories to inputs to expansions */
	public static Map<String, Map<String, String>> headMap = new HashMap<String, Map<String, String>>();
	
    /** For training. Constructs a BinaryTree with category <tt>c</tt>.
	 */
	public BinaryTree(String c){	
		this.cat = c;	
		this.head = null;
		this.sister = null;
	}

	public abstract BinaryTree copy();
	
	public abstract BinaryTree deepCopy();

	public String lexCat(){
		if (this instanceof BinaryNode){
			return head.lexCat();
		}
		else 
			return cat;	  
	}

	/** returns <tt>cat</tt>.
	 *  @see BinaryTree#cat
	 */
	public String cat(){
		return cat;
	}

	/** returns <tt>headpos</tt>. 
	 * @see BinaryTree#headpos
	 */
	public int headPos(){
		return headpos;
	}

	/** returns the expansion <tt>exp</tt> as string */
	public String exp(){     
		return exp;
	}
	
	/** <pre>exp|cat</pre> */
	public String expCat(){
		return (new StringBuffer(exp()).append('|').append(lexCat())).toString();
	}
	
	public int headCatInt(){
		return -1;
	}

	/** @see BinaryNode#sisterCat()
	 */
	public String sisterCat(){     
		return new String();
	}
	public int sisterCatInt(){     
		return -1;
	}

	/** can be used for unknown words: true if the first character is a digit */
	public static boolean isNumber(String word){    
		if (word != null 
				&& word.length() > 0 
				&& Character.isDigit(word.charAt(0)))
			return true;
		else return false;
	}

	/** can be used for unknown words: true if the first character is upper case */ 
	public static boolean isUpperCase(String word){
		if (word != null && word.length() > 0 && Character.isUpperCase(word.charAt(0)))
			return true;
		else return false;
	}

	public static void addEntireTree(String line){
		StringBuffer tmp = new StringBuffer(line);
		// set this to where you're currently in line
		int currentInd = 0;	
		int oBInd = line.indexOf('(', currentInd);
		int cArrow = line.indexOf('>', currentInd);
		int cBInd = line.indexOf(')', currentInd);
		int length = line.length();
		String token; 

		while (currentInd < length){
			oBInd = line.indexOf('(', currentInd) +1;
			cArrow = line.indexOf('>', currentInd) + 1;
			cBInd = line.indexOf(')', currentInd) + 1;
			// case 1: token is "(<...>"
			// get everything from ( to >
			if (oBInd < cArrow && oBInd < cBInd){
				token = tmp.substring(currentInd, cArrow); 
				currentInd = cArrow;
				while (currentInd < length && tmp.charAt(currentInd) == ' '){
					currentInd++;
				}	
			}
			else{
				token = tmp.substring(currentInd, cBInd);
				currentInd = cBInd+1;
				while (currentInd < length && tmp.charAt(currentInd) == ' '){
					currentInd++; 
				}
			}
			addToken(token);
		}
	}

	/** Used in training: calls <tt>addChild()</tt> 
	if the String token starts a new constituent, 
	or <tt>endConstit()</tt> if it ends a constituent. 
       @see BinaryTree#isStartConstit
       @see BinaryTree#addChild(String)
       @see BinaryTree#isEndConstit
       @see BinaryTree#endConstit()*/
	public static void addToken(String token){
		if (isStartConstit(token)){
			addChild(description(token));
		}
		if (isEndConstit(token)){
			endConstit();
			
			String key;
			if ((current.head != null)){
				String resultCat = current.cat;
				//Ignore TOP unary productions
				if (resultCat.equals("TOP")) return;
				String headChild = current.head.cat;
				String resultExp = current.exp;
				key = headChild;
				if (resultExp != unaryExp){
					if (current.sister == null) return;
					String sisterChild = current.sister.cat;
					if (resultExp == leftHeadExp) key = headChild+"+"+sisterChild;
					else key = sisterChild+"+"+headChild;
				}
				if (!headMap.containsKey(resultCat)){
					Map<String, String> inputs = new HashMap<String, String>();
					inputs.put(key, resultExp);
					headMap.put(resultCat, inputs);
				}
				else {
					if (!headMap.get(resultCat).containsKey(key))
						headMap.get(resultCat).put(key, resultExp);
				}
			}
		}
	}
	
	public static Map<String, Map<String, String>> headMap(){return headMap;}

	/** adjusts delta (why?) and sets current to its parent */
	public static void endConstit(){
		if (current != null && current.parent != null){
			current = current.parent;
		}
	}

	/** Returns the actual description of the node, cutting out the initial "(" */
	public static String description(String token){
		StringTokenizer tokenizer = new StringTokenizer(token.substring(1), "<>");
		String des = "";
		if (tokenizer.hasMoreTokens())
			des = tokenizer.nextToken();	
		return des;
	}

	/** adds a new child to the tree <tt>current</tt>.
	Called by <tt>addToken</tt>.
     @see BinaryTree#current
     @see BinaryTree#addToken
     @see BinaryTree#description
	 */
	public static void addChild(String description){
		boolean leaf = false;
		String cat = new String();
		String headword = new String();
		String label = new String();
		int dtrs_nr = 0;
		int childHead = -1; // == exp!!!
		String exp = ""; // meaning: neither left, right, nor unary (nor coordination)
		BinaryTree child;
		StringTokenizer tokenlist = new StringTokenizer(description);

		// (1) Leaf or internal node?
		if (tokenlist.hasMoreTokens()){
			if ( (tokenlist.nextToken()).equals("L"))
				leaf = true;
			else leaf = false;
		}

		// (3) Category -- IN BOTH FORMATS
		if (tokenlist.hasMoreTokens()){	    
			if (!leaf){
				cat = (new StringBuffer(tokenlist.nextToken())).toString();	
			}
			else {
				cat = tokenlist.nextToken();
			}
		}

		// (4) Leaf ==> word
		if (leaf){
			// tag: (in both formats)
			if (tokenlist.hasMoreTokens())
				label = tokenlist.nextToken();	
			// next is either original tag followed by head word (in new format) or head word (in old format):
			if (tokenlist.hasMoreTokens()){
				String next = tokenlist.nextToken();	
				String tmpWord = next;
				//System.out.println("next: " + next);
				if (tokenlist.hasMoreTokens()){
					tmpWord = tokenlist.nextToken();
				}	
				headword = tmpWord;  
			}
			exp = leafExp;
		}
		// (4) Internal node:  (label) + head + dtrs_nr 
		// (this determines the expansion)
		else{ 
			if (tokenlist.hasMoreTokens()){
				String labelOrHead = tokenlist.nextToken();		
				if (Character.isDigit(labelOrHead.charAt(0))){
					childHead = Integer.parseInt(labelOrHead);
				}
				else{
					label = labelOrHead;		    
					if (tokenlist.hasMoreTokens()){
						childHead = Integer.parseInt(tokenlist.nextToken()); 
					}
				}
			}
			if (tokenlist.hasMoreTokens()){
				dtrs_nr = Integer.parseInt(tokenlist.nextToken()); 	
			}
			if (dtrs_nr == 1) exp = unaryExp;
			else exp = convertExp(childHead);	
		}

		// (5)  Create the new constituent
		if (leaf){   
			child = new BinaryLeaf(cat, headword, label);	    	   
		}
		else { 
			child = new BinaryNode(cat, exp);
		}

		// (6) Insert child into tree
		child.parent = current;	       
		if (current.head != null && current.exp != unaryExp){	
			current.sister = child;	    

		} 
		else{
			if (current.sister!= null){	   
				current.head = child;		
			}
			else {// first child!		
				if ( current.exp == leftHeadExp || current.exp == unaryExp){		
					current.head = child;
				}
				else {
					current.sister = child;
				}
			}
		}
		current = child;		
	}
	
	private static String convertExp(int exp){
		//0: left
		//1: right
		//2: unary 
		//3: leaf 
		String convExp="";
		switch(exp){
			case 0: convExp=leftHeadExp; break;
			case 1: convExp=rightHeadExp; break;
			case 2: convExp=unaryExp; break;
			case 3: convExp=leafExp; break;
		}
		return convExp;
	}

	/** True if token starts with open brackets 
     @see BinaryTree#addChild
	 */
	private static boolean isStartConstit(String token){	
		return token.startsWith(open_brackets);
	}
	/** True if token ends with closed brackets 
	 @see BinaryTree#addChild
	 */
	private static boolean isEndConstit(String token){	
		return token.endsWith(closed_brackets);
	}
	
	public abstract String toString();
	
	protected abstract int headIndex();
	protected abstract String headWord();
	protected abstract String headCat();
	
	public PargStructure[] toHeadDeps(PargStructure[] heads){
		StringUtils s = new StringUtils();
		if (exp.equals(leafExp)) return heads;
		if (exp.equals(unaryExp)) return head.toHeadDeps(heads);
		BinaryTree tempSis, tempHead;
		tempSis = sister;
		tempHead = head;
		int sisIndex = tempSis.headIndex();
		String sisWord = tempSis.headWord();
		int headIndex = tempHead.headIndex();
		String headWord = tempHead.headWord();
		if (!s.isPunct(sisWord) && !s.isPunct(headWord)){
			String headCat = tempHead.headCat();
			if (heads[sisIndex] == null) heads[sisIndex] = new PargStructure(sisWord);
			heads[sisIndex].addHead(headIndex, headWord, headCat);
		}
		heads = head.toHeadDeps(heads);
		heads = sister.toHeadDeps(heads);
		return heads;
	}
}

