//////////////////////////////////////////////////////////////////////////////
//Copyright (C) 2011-12 Tejaswini Deoskar

//This library is free software; you can redistribute it and/or
//modify it under the terms of the GNU Lesser General Public
//License as published by the Free Software Foundation; either
//version 2.1 of the License, or (at your option) any later version.

//This library is distributed in the hope that it will be useful,
//but WITHOUT ANY WARRANTY; without even the implied warranty of
//MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//GNU Lesser General Public License for more details.

//You should have received a copy of the GNU Lesser General Public
//License along with this program; if not, write to the Free Software
//Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

package edinburgh.ccg.probModel.util;


import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.StringTokenizer;

import edinburgh.ccg.util.CCGbankStyleWordPOSTokeniser;
import edinburgh.ccg.util.Logger;
import edinburgh.ccg.util.Options;


/**
 * Counts lexical frequencies in training data. 
 * Used for smoothing the lexicon. 
 * Based on STLexCreator
 * Can be used to smooth lexicon and include new words, instead of using the Supertagger
 * @author Tejaswini Deoskar
 *
 */
public class SmoothLexCreator {

	private Options o = Options.getOptions();
	private final String trainFileName = o.getTrainFile();
	private Logger log;

	private File trainFile;

	private int CUT_OFF = o.getSmoothCutOff(); 

	private HashMap<String, HashMap <String, Double>> testWordTagFreq = new HashMap<String, HashMap <String, Double>> ();
	private HashMap<String, Integer> testTagFreq = new HashMap<String, Integer> ();

	//This Hash stores the counts of words and categories from test data. 
	//The keys are in format 'word|cat'
	private HashMap< String, Double> testWordCatFreq = new HashMap < String, Double> ();

	//Hashes to store counts from training data. 
	private HashMap <String, HashMap <String, Double>> trainPOSCatFreq = new HashMap <String, HashMap <String, Double>>();;
	private HashMap <String, Double> trainPOSFreq = new HashMap <String, Double> ();;
	private HashMap < String, HashMap <String, Double>> trainWordCatFreq =  new HashMap < String, HashMap <String, Double>> ();;
	private HashMap <String, Integer> trainCatFreq= new HashMap <String, Integer> (); //Cat = CCG Category
	//private HashMap <String, Integer> trainWordFreq = new HashMap <String, Integer> ();
	private HashMap <String,Integer> wordFreq  = new HashMap<String,Integer>();

	//This hash saves the train lexicon as word|cat and frequency. Contains the same info 
	// as the Hash of Hashes 'trainWordCatFreq', but this is a simple Hash.
	public HashMap <String, Double> trainWordCatFreqHash= new HashMap <String, Double> ();

	public HashMap <String, HashMap <String,Double>> trainWordTagFreq= new HashMap <String, HashMap <String,Double>> ();

	public HashMap <String, HashMap <String,Double>> testWordCatFreq1  =  new HashMap < String, HashMap <String, Double>> ();
	//This Hash saves the smoothed lexicon as word|cat and frequency.
	public HashMap <String, Double> smoothedWordCatFreq= new HashMap <String, Double> ();

	public SmoothLexCreator(Logger log){
		trainFile = new File(trainFileName);
		this.log = log;
		log.println("\tMerging test words into model and smoothing...");
		createTrainLex();
		log.println("\tdone!");
	}


	private void createTrainLex(){
		HashMap < String, List <String> > wordCats = new HashMap<String,List<String>>();
		//HashMap<String,Integer> wordFreq = new HashMap<String,Integer>();
		HashMap<String,List<String>> posTags = new HashMap<String,List<String>>();

		String line = null;
		BufferedReader trainingBuffer = null;
		CCGbankStyleWordPOSTokeniser tok = new CCGbankStyleWordPOSTokeniser();
		List<String[]> tempTokens;
		List<String> cats;

		// Read the training file
		try {
			trainingBuffer = new BufferedReader(new FileReader(trainFile));
		}
		catch (FileNotFoundException e){
			System.err.println("ERROR: Unable to read training file: "+trainFile);
			System.exit(0);
		}
		try {
			line = trainingBuffer.readLine();
			while (line != null && line.indexOf("ID=") != -1){
				line = trainingBuffer.readLine();
				tempTokens = tok.getTokenGroups(line);
				for (String[] tokenGroup : tempTokens){
					// CCG cat = tokenGroup[0]
					// POS-tag = tokenGroup[1]
					// word = tokenGroup[3]
					String tempWord = tokenGroup[3];
					String tempCat = tokenGroup[0];
					String tempPOS = tokenGroup[1];

					// Update the word frequency counts
					if (wordFreq.containsKey(tempWord)){
						int freq = ((Integer)wordFreq.get(tempWord)).intValue();
						wordFreq.put(tempWord, new Integer(++freq));
					}
					else  wordFreq.put(tempWord, new Integer(1));
					// Update the category list for this word
					cats = wordCats.get(tempWord);
					if (cats == null) cats = new ArrayList<String>();
					if (!cats.contains(tempCat)) cats.add(tempCat);
					wordCats.put(tempWord, cats);
					// Update the category list for this POS-tag
					cats = posTags.get(tempPOS);
					if (cats == null) cats = new ArrayList<String>();
					if (!cats.contains(tempCat)) cats.add(tempCat);
					posTags.put(tempPOS, cats);

					//Fill Cat -> Freq (Cat = CCG Category)
					if (trainCatFreq.containsKey(tempCat)){
						Integer freq = (Integer) trainCatFreq.get(tempCat);
						++freq; 
						trainCatFreq.put(tempCat, freq);
					}
					else trainCatFreq.put(tempCat, new Integer (1));

					//Fill POS -> Freq
					if (trainPOSFreq.containsKey(tempPOS)){
						Double freq = (Double) trainPOSFreq.get(tempPOS);
						++freq; 
						trainPOSFreq.put(tempPOS, freq);
					}
					else trainPOSFreq.put(tempPOS, new Double (1.0));

					//Fill POS -> CAT -> Freq
					if(trainPOSCatFreq.containsKey(tempPOS)){
						HashMap<String, Double> m = trainPOSCatFreq.get(tempPOS);
						if(m.containsKey(tempCat)){
							Double freq1 = (Double) m.get(tempCat);
							++freq1;
							m.put(tempCat, freq1);
						}
						else m.put(tempCat,new Double (1.0)); 

						trainPOSCatFreq.put(tempPOS, m);
					}
					else {
						HashMap<String, Double> m = new HashMap<String, Double> ();
						m.put(tempCat, new Double (1.0));
						trainPOSCatFreq.put(tempPOS, m);
					}

					//Fill Word -> Cat -> Freq
					if(trainWordCatFreq.containsKey(tempWord)){
						HashMap<String, Double> m = trainWordCatFreq.get(tempWord);
						if(m.containsKey(tempCat)){
							Double freq1 = (Double) m.get(tempCat);
							++freq1;
							m.put(tempCat, freq1);
						}
						else m.put(tempCat,new Double (1.0)); 

						trainWordCatFreq.put(tempWord, m);
					}
					else {
						HashMap<String, Double> m = new HashMap<String, Double> ();
						m.put(tempCat, new Double (1.0));
						trainWordCatFreq.put(tempWord, m);
					}

					//Fill Word -> POS -> Freq
					if(trainWordTagFreq.containsKey(tempWord)){
						HashMap<String, Double> m = trainWordTagFreq.get(tempWord);
						if(m.containsKey(tempPOS)){
							Double freq1 = (Double) m.get(tempPOS);
							++freq1;
							m.put(tempPOS, freq1);
						}
						else m.put(tempPOS,new Double (1.0)); 

						trainWordTagFreq.put(tempWord, m);
					}
					else {
						HashMap<String, Double> m = new HashMap<String, Double> ();
						m.put(tempPOS, new Double (1.0));
						trainWordTagFreq.put(tempWord, m);
					}

					//Fill Word|Cat -> Freq
					if(trainWordCatFreqHash.containsKey(tempWord+"|"+tempCat)){
						Double freq = (Double) trainWordCatFreqHash.get(tempWord+"|"+tempCat);
						++freq; 
						trainWordCatFreqHash.put(tempWord+"|"+tempCat, freq);
					}
					else {
						trainWordCatFreqHash.put(tempWord+"|"+tempCat, new Double (1.0));
					}

				}
				line = trainingBuffer.readLine();
			}
		}
		catch(IOException e) { e.printStackTrace();}
	}

	public void updateTestCounts (String sentence ) {
		StringTokenizer tokenizer = new StringTokenizer(sentence);

		if (tokenizer.hasMoreTokens()){
			String  next = tokenizer.nextToken();
			String[] tmp = next.split("_");
			String tempWord = new String(tmp[0]);
			String tempPOS = new String(tmp[1]);

			//Fill Test Hashes 
			//Fill POS -> Freq
			if (testTagFreq.containsKey(tempPOS)){
				Integer freq = (Integer) testTagFreq.get(tempPOS);
				++freq; 
				testTagFreq.put(tempPOS, freq);
			}
			else testTagFreq.put(tempPOS, new Integer (1) );

			//Fill Word -> POS -> Freq
			if(testWordTagFreq.containsKey(tempWord)){
				HashMap<String, Double> m = testWordTagFreq.get(tempWord);
				if(m.containsKey(tempPOS)){
					Double freq1 = (Double) m.get(tempPOS);
					++freq1;
					m.put(tempPOS, freq1);
				}
				else m.put(tempPOS,new Double (1.0)); 

				testWordTagFreq.put(tempWord, m);
			}
			else {
				HashMap<String, Double> m = new HashMap<String, Double> ();
				m.put(tempPOS, new Double (1.0));
				testWordTagFreq.put(tempWord, m);
			}

		}
		while (tokenizer.hasMoreTokens()){
			String  next = tokenizer.nextToken();				

			String[] tmp = next.split("_");
			String tempWord = new String(tmp[0]);
			String tempPOS = new String(tmp[1]);

			//Fill POS -> Freq
			if (testTagFreq.containsKey(tempPOS)){
				Integer freq = (Integer) testTagFreq.get(tempPOS);
				++freq; 
				testTagFreq.put(tempPOS, freq);
			}
			else testTagFreq.put(tempPOS, new Integer (1) );

			//Fill Word -> POS -> Freq
			if(testWordTagFreq.containsKey(tempWord)){
				HashMap<String, Double> m = testWordTagFreq.get(tempWord);
				if(m.containsKey(tempPOS)){
					Double freq1 = (Double) m.get(tempPOS);
					++freq1;
					m.put(tempPOS, freq1);
				}
				else m.put(tempPOS,new Double (1.0)); 

				testWordTagFreq.put(tempWord, m);
			}
			else {
				HashMap<String, Double> m = new HashMap<String, Double> ();
				m.put(tempPOS, new Double (1.0));
				testWordTagFreq.put(tempWord, m);
			}
		}
	}


	public void smooth(){
		Double lambda = o.getSmoothLambda(); //This is the interpolation factor for the test corpus. 
		Double smoothedFreq = new Double (0.1);		
		//First, assign Cats to POS-tagged words in test data
		assignCatsTest(); //This is a hash with test words containing all categories associated
		//with their POS tags.
		//Should do this only for words above CUT-OFF in train. 

		//Create a new Hash of all words: test and train
		HashMap<String, HashMap <String, Double>> tmpTrainTest = new HashMap<String, HashMap <String, Double>> ();
		//Add all words from train and test into this hash: the values don't matter.
		tmpTrainTest.putAll(trainWordTagFreq);
		tmpTrainTest.putAll(testWordTagFreq); //Could also be testWordCatFreq1

		//Iterate over all words in train and test
		for (Map.Entry< String, HashMap<String, Double>> word : tmpTrainTest.entrySet() ) {
			HashMap<String, Double> tags = new HashMap<String, Double>();
			tags = word.getValue();

			//Iterate over all POS tags of the word, either in train or test. 
			for (Map.Entry <String, Double> tag : tags.entrySet()){
				//For all categories in trainPOSCatFreq

				//Map to store categories of each POS tags from train data.
				HashMap<String, Double> cats = new HashMap<String, Double>();		
				cats=trainPOSCatFreq.get(tag.getKey());

				if(cats != null) { //Only if there are categories associate with the POS tag, else skip

					//for all categories of the POS tag from train data
					for (Map.Entry <String, Double> cat : cats.entrySet()) {
						String tempWordCat = new String (word.getKey() + "|" + cat.getKey() );
						//System.out.println("All Word-CAT combinations (train and test): "+tempWordCat );
						HashMap<String, Double> tmpTestHash =  new HashMap<String, Double>();
						tmpTestHash = testWordTagFreq.get(word.getKey()) ;

						HashMap<String, Double> tmpTrainHash =  new HashMap<String, Double>();
						tmpTrainHash = trainWordTagFreq.get(word.getKey()) ;

						//If word exists in test, does not exist in train (that is, unseen word)
						if(tmpTestHash != null && tmpTestHash.containsKey(tag.getKey()) && tmpTrainHash == null ) {
							smoothedFreq =  lambda * testWordCatFreq.get(tempWordCat);
							smoothedWordCatFreq.put(tempWordCat, smoothedFreq );
						}									

						//If word exists in test and train, but tag does not exist in train
						//that is, a word is tagged with an unseen  tag in the test corpus
						if(tmpTestHash != null && tmpTestHash.containsKey(tag.getKey()) && tmpTrainHash != null ) {

							if(tmpTrainHash.containsKey(tag.getKey()) != true) {
								smoothedFreq =  lambda * testWordCatFreq.get(tempWordCat);
								smoothedWordCatFreq.put(tempWordCat, smoothedFreq );
							}
						}

						//If (word, tag) exists in test and train and is below CUT_OFF Freq in train
						if(tmpTestHash != null && tmpTestHash.containsKey(tag.getKey())&&  tmpTrainHash != null && tmpTrainHash.containsKey(tag.getKey()) ) { 
							if( tmpTrainHash.get(tag.getKey()) < CUT_OFF) {
								//Even for words to be smoothed (i.e. below cut-off, can use train frequencies 
								//only for categories that have been seen with that word in train data
								if (trainWordCatFreqHash.get(tempWordCat) !=null){
									smoothedFreq = (1 - lambda) *trainWordCatFreqHash.get(tempWordCat) + lambda * testWordCatFreq.get(tempWordCat);
								}
								//Else use only the second term 
								else { smoothedFreq = lambda * testWordCatFreq.get(tempWordCat); }
								smoothedWordCatFreq.put(tempWordCat, smoothedFreq );
							}
							else {
								if (trainWordCatFreqHash.get(tempWordCat) !=null){
									smoothedFreq = (1 - lambda) *trainWordCatFreqHash.get(tempWordCat) ;
									smoothedWordCatFreq.put(tempWordCat, smoothedFreq );
								}
							}
						}
						//If word exists in train but not in test
						if(tmpTestHash == null && tmpTrainHash != null){
							if(trainWordCatFreqHash.get(tempWordCat) !=null){
								smoothedFreq = (1 - lambda) * trainWordCatFreqHash.get(tempWordCat) ;
								smoothedWordCatFreq.put(tempWordCat, smoothedFreq);
							}
						}
					}
				}
				else {log.println("Warning: No categories for " + tag.getKey() ); }
			}		
		}
	}
	
	
	public void printSmoothedLex (String dir, String filename){
		BufferedWriter out;
		try {
			out = new BufferedWriter(new FileWriter (dir+"/"+filename));
			//out.write("Train POS Frequencies:");
			
			
			//out.newLine();
			for (Map.Entry<String, Double> entry :smoothedWordCatFreq.entrySet() ) {		
			    //System.out.println(entry.getKey() + ": " + entry.getValue());
					out.write(entry.getKey()+" "+entry.getValue().toString());
					out.newLine();
				}
			out.close();
		}
		catch (IOException e){e.printStackTrace();}
		
	}
	

	public void assignCatsTest(){
		//Iterate over all test words 
		for (Map.Entry<String, HashMap<String, Double>> entry : testWordTagFreq.entrySet() ) {

			HashMap<String, Double> m = new HashMap<String, Double>();
			m = entry.getValue();

			String tempWord = new String (entry.getKey().toString());

			//Iterate over all POS tags of the word in test data
			for (Map.Entry<String, Double> entry1 : m.entrySet()) {

				if (entry1.getValue() != null){   //c(w,T)

					Double tempNum = new Double (0.0);
					Integer tempDeno = new Integer (1);

					String tempPOS = new String (entry1.getKey().toString());		
					tempNum  = trainPOSFreq.get(tempPOS); //t(T)
					tempDeno = testTagFreq.get(tempPOS);  //c(T) 

					Double scaled_freq = new Double(0.0);

					//Scale test word counts by treebank counts
					if(tempNum != null  && tempDeno != 0) {	
						scaled_freq = tempNum / tempDeno * entry1.getValue(); //c'(w,T) 
					}
					else { 
						log.println("Warning: Frequency of POS tag is zero in training data: " + tempPOS); 
					}

					/* For a POS tag (present in test data), do for all categories that have occurred 
					// with that POS. i.e. divide freq of POS tag into all categories 
					// associated with that tag according to some ratio. 
					*/

					if(trainPOSCatFreq.containsKey(tempPOS)){
						HashMap<String, Double> mm = new HashMap<String, Double>();
						mm = trainPOSCatFreq.get(tempPOS);

						for (Map.Entry < String, Double> catentry : mm.entrySet()){
							String tempCat = new String (catentry.getKey().toString());
							Double tempFreq = new Double(0.0);
							tempFreq = mm.get(tempCat) / trainPOSFreq.get(tempPOS) *scaled_freq ;
							testWordCatFreq.put(tempWord +"|"+ tempCat, tempFreq);
							HashMap < String, Double> nn = new HashMap < String, Double>();
							nn.put(tempCat, tempFreq);
							testWordCatFreq1.put(tempWord, nn);
						}			
					}	
					else {// IF POS tag does not have any categories, then something 
						//else needs to be done!											
					}
				} 
			}		
		} //testWordCatFreq Hash is populated now. 
	}
}
