/*
 * Copyright (c) 2007, Dennis N. Mehay
 * CCGbankStyleWordPOSTokeniser.java
 *
 * Created on October 24, 2007, 2:19 PM
 */

package edinburgh.ccg.util;

import java.util.ArrayList;
import java.util.NoSuchElementException;
import java.util.StringTokenizer;
import java.util.regex.Matcher;
import java.util.regex.Pattern;
import java.util.regex.PatternSyntaxException;

/**
 *
 * @author Dennis N. Mehay
 */
public class CCGbankStyleWordPOSTokeniser {
	private static Pattern p = Pattern.compile("(<L\\s+.*?>)+?");

	/** Creates a new instance of CCGbankStyleWordPOSTokeniser */
	public CCGbankStyleWordPOSTokeniser() {
	}

	/** 
	 * A method to step through a space-delimited (i.e., space-*tokenised*) <code>String</code> 
	 * representing a CCGbank-style lexical entry of the form <L CCG-cat POS1 POS2 word CCG-cat-w-deps>, 
	 * getting all the relevant tokens therein (just CCG-cat, POS1 and word for now).  
	 * (N.B. This method does NOT check to see whether the <code>String</code> you give it is of the 
	 * proper form.
	 *
	 * @param lex A <code>String</code> representing a lexical entry in the CCGbank.
	 * @return A <code>String[]</code> containing the tokens (excluding the `<L' and `>'
	 *         CCGbank delimiting tokens) in the CCGbank-style lexical entry.
	 */
	public String[] getTokens(String lex) {
		StringTokenizer st = new StringTokenizer(lex);
		ArrayList<String> temp = new ArrayList<String>();
		try {
			st.nextToken();           // Throw away the `<L'.
			temp.add(st.nextToken()); // The CCG category.
			temp.add(st.nextToken()); // The Penn Treebank POS tag (corrected).
			temp.add(st.nextToken()); // The uncorrected P.Treebank POS tag.
			temp.add(st.nextToken()); // The word.
			String nextTok = st.nextToken();

			// See whether the next token has > in it or not.
			// If so, it's the last token, otherwise, loop till the end.
			while(!nextTok.contains(">")) {
				temp.add(nextTok);
				nextTok = st.nextToken();
			}
			// OK, this is the last token.
			String[] lastOne = (nextTok).split(">", 0);
			temp.add(lastOne[0]); // Usually the CCG category with indices and dependency info.
			String[] retVal = new String[temp.size()];
			for(int i = 0; i<temp.size(); i++) {
				retVal[i] = temp.get(i);
			}
			return retVal;
		}
		catch(NoSuchElementException nsee) { 
			System.out.println("This is not a CCGbank-style lexical item of the form\n" + 
			"<L CCG-cat POS1 POS2 word CCG-cat-w-dep ...>.");
			return null;
		}
		catch(NullPointerException npe) {
			System.out.println("The String passed in was null.");
			return null;
		}
	} 

	/** 
     * A method that collects all lexical items (i.e., lexical nodes) from a 
     * line of text containing a CCGbank-style parse.
     *
    */
    @SuppressWarnings("static-access")
	public ArrayList<String> getLexicalItems(String line, boolean getPunctuationNodes) {
		ArrayList<String> retVal = new ArrayList<String>();
		try {
			Matcher m = this.p.matcher(line);

			while(m.find()) {
				retVal.add(m.group());
			}
			return retVal;
		}
		catch(IllegalStateException ise) { 
			System.out.println("Line: " + line + "\n" +
					"must not be a CCGbank-style parse.\n" +
			"The ArrayList<String> being returned is empty.");
			return retVal;
		}
		catch(PatternSyntaxException pse) { return retVal; } // This will never happen. Our pattern is legal.

	} // End method getLexicalItems   

	public ArrayList<String[]> getTokenGroups(String line) {
		ArrayList<String[]> retVal;
		ArrayList<String> temp = this.getLexicalItems(line, true);
		retVal = new ArrayList<String[]>(temp.size());
		for(int i=0; i<temp.size(); i++) {
			retVal.add(this.getTokens(temp.get(i)));
		}
		return retVal;
	}
}

