//////////////////////////////////////////////////////////////////////////////
//Copyright (C) 2011 Christos Christodoulopoulos

//This library is free software; you can redistribute it and/or
//modify it under the terms of the GNU Lesser General Public
//License as published by the Free Software Foundation; either
//version 2.1 of the License, or (at your option) any later version.

//This library is distributed in the hope that it will be useful,
//but WITHOUT ANY WARRANTY; without even the implied warranty of
//MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//GNU Lesser General Public License for more details.

//You should have received a copy of the GNU Lesser General Public
//License along with this program; if not, write to the Free Software
//Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

package edinburgh.ccg.util;

import java.io.File;


/**
 * A storage class to keep all the sampler options.
 * 
 * @author Christos Christodoulopoulos
 */
public class Options {
	private boolean beamSearch;
	private boolean train;
	private boolean smoothTrain; 
	private double smoothLambda = 0.01; 
	private int smoothCutOff = 1;
	private boolean backoff;
	private boolean useTags;
	/** This is linked to {@link timeoutMins}. */
	private boolean timeout;
	/** Time (in minutes) before timeout for each parse. */
	private double timeoutMins;
	/** Threshold for rare categories (during training). */
	private int catThreshold = 0;
	private int rareWordThreshold = 5;
	private String parseFile, smoothTrainFile;
	
	/** Used for checking the CCG-legality of a unification */
	private boolean ccgCheck = false;
	
	private String modelsDir = new File("models").getAbsolutePath();
	private String trainFile;
	private String outputFile = "parses.txt";
	
	private final String wordProbModelFileName = "wordProbModel";
	private final String expProbModelFileName = "expProbModel";
	private final String headCatProbModelFileName = "headCatProbModel";
	private final String sisterCatProbModelFileName = "sisterCatProbModel";
	private final String sisterLexCatProbModelFileName = "sisterLexCatProbModel";
	private final String headMapFileName = "headMap";
	
	/** The beam width */
	private double alpha = 0.0001;
	/** The statistical backoff probability (set this as low as possible)*/
	private final double backoffProb = 10^(-10);
	
	/** The ProbModel type **/
	private String modelType = "base";
	
	/** The static object for passing the options around (using a getter). */
	private static Options theOptions;
	
	/**
	 * Constructor for the graphical interface.
	 * To be used with setters.
	 */
	public Options(){
		theOptions = this;
	}

	/**
	 * Constructor for the command-line interface.
	 * Reads from stdin.
	 * @param args The command-line parameters
	 */
	public Options(String[] args){
		theOptions = this;
		for (int i = 0; i < args.length; i++) {
			
			if (args[i].equals("-train") || args[i].equals("-t")){
				train = true;
				if (!checkNext(args,i)){
					trainFile = args[++i];
				}
			}
			else if (args[i].equals("-models") || args[i].equals("-m")){
				if (checkNext(args,i)){
					System.err.println("No model directory given!");
					System.exit(1);
				}
				else modelsDir = args[++i];
			}
			else if (args[i].equals("-mt")){
				if (checkNext(args,i)){
					System.err.println("No model type given! Using baseline model by default.");
					System.exit(1);
				}
				else modelType = args[++i];
			}
			else if (args[i].equals("-smooth") || args[i].equals("-s")){
				smoothTrain = true;
				if (checkNext(args,i)){
					System.err.println("No smoothing file given!");
					System.exit(1);
				}
				else smoothTrainFile = args[++i];
			}
			else if (args[i].equals("-smoothfactor") || args[i].equals("-sf")){
				smoothTrain = true;
				if (checkNext(args,i)){
					System.err.println("No smoothing factor given!");
					System.exit(1);
				}
				else smoothLambda = Double.parseDouble(args[++i]);
			}
			else if (args[i].equals("-smoothcutoff") || args[i].equals("-sc")){
				smoothTrain = true;
				if (checkNext(args,i)){
					System.err.println("No smoothing cutoff given!");
					System.exit(1);
				}
				else smoothCutOff = Integer.parseInt(args[++i]);
			}
			else if (args[i].equals("-output") || args[i].equals("-o")){
				if (!checkNext(args,i)){
					outputFile = args[++i];
				}
			}
			else if (args[i].equals("-beam") || args[i].equals("-b")){
				beamSearch = true;
				if (!checkNext(args,i)){
					alpha = Double.parseDouble(args[++i]);
				}
			}
			else if (args[i].equals("-backoff") || args[i].equals("-bo")){
				backoff = true;
			}
			else if (args[i].equals("-pos")){
				useTags = true;
			}
			else if (args[i].equals("-ccg")){
				ccgCheck = true;
			}
			else if (args[i].equals("-timeout") || args[i].equals("-to")){
				timeout = true;
				if (checkNext(args,i)){
					System.err.println("No timout value given!");
					System.exit(1);
				}
				else timeoutMins = Double.parseDouble(args[++i]);
			}
			else if (args[i].equals("-catthreshold") || args[i].equals("-c")){
				//RETRAIN THE MODELS IF THIS IS CHANGED
				train = true;
				if (checkNext(args,i)){
					System.err.println("No category threshold value given!");
					System.exit(1);
				}
				else catThreshold = Integer.parseInt(args[++i]);
			}
			else {
				parseFile = args[i];
			}
		}		
	}
	
	private boolean checkNext(String[] args, int i){
		if (args.length <= i+1) return true;
		if (args[i+1].startsWith("-")) return true;
		return false;
	}
	
	public String toString(){
		String text = "### Using settings:\n";
		text += "#Models directory:\t"+modelsDir+"\n";
		text +=	"#Train file:\t\t"+trainFile+"\n";
		if (smoothTrain) text += "#Smoothed tr. file:\t"+smoothTrainFile+"\n";
		if (smoothTrain) text += "#Smoothing cut-off:\t"+smoothCutOff+"\n";
		if (smoothTrain) text += "#Smoothing interp. factor:\t"+smoothLambda+"\n";
		text += "#File to parse:\t\t"+parseFile+"\n";
		text += "#Output file:\t\t"+((outputFile!=null) ? outputFile : "stdout")+"\n";
		text += "#Beam search:\t\t"+(beamSearch ? "on" : "off")+"\n";
		if (beamSearch) text += "#Beam width:\t\t"+alpha+"\n";
		text += "#Simple back-off:\t"+(backoff ? "on" : "off")+"\n";
		text += "#Back-off to PoS tags:\t"+(useTags ? "on" : "off")+"\n";
		text += "#Timeout:\t\t"+(timeout ? "on" : "off")+"\n";
		if (timeout) text += "#Timeout time (mins):\t"+timeoutMins+"\n";
		text += "#Rare cat threshold:\t"+catThreshold+"\n";
		if (catThreshold!=0) text += "#(Please retrain if changed)\n";
		return text;
	}
	
	public static Options getOptions(){
		//Check if initialised
		return theOptions;
	}
	
	public void setModelsDir(String dir){
		this.modelsDir = dir;
	}
	public void setParseFile(String parseFile){
		this.parseFile = parseFile;
	}
	public void setTrainFile(String trainFile) {
		this.trainFile = trainFile;
	}
	public void setSmoothFile(String smoothFile) {
		this.smoothTrain = true;
		this.smoothTrainFile = smoothFile;
	}
	public void setOutputFile(String outFile) {
		this.outputFile = outFile;
	}
	public void setBeamSearch(boolean beam) {
		this.beamSearch = beam;
	}
	public void setAlpha(double alpha) {
		this.alpha = alpha;
	}
	public void setSmoothCutOff(int cutoff) {
		this.smoothCutOff = cutoff;
	}
	public void setSmoothInterpFactor(double factor) {
		this.smoothLambda = factor;
	}
	public void setRetrain(boolean retrain) {
		this.train = retrain;
	}
	public void setUseTags(boolean pos){
		this.useTags = pos;
	}
	public void setModelType(String type){
		modelType = type;
	}
	
	public boolean isBeamSearch() {return beamSearch;}
	public boolean isTrain() {return train;}
	public boolean isSmoothTrain() {return smoothTrain;}
	public boolean isTimeout() {return timeout;}
	public boolean isBackoff() {return backoff;}
	public boolean useTags() {return useTags;}
	public double getTimeoutMins() {return timeoutMins;}
	public int getCatThreshold() {return catThreshold;}
	public int getRareWordThreshold() {return rareWordThreshold;}
	public String getModelsDir() {
		File mF = new File(modelsDir);
		if (!mF.exists()) mF.mkdir();
		return modelsDir;
	}
	public String getOutputFile() {return outputFile;}
	public String getParseFile() {return parseFile;}
	public String getTrainFile() {return trainFile;}
	public String getSmoothTrainFile() {return smoothTrainFile;}
	public String getWordProbModelFile() {return wordProbModelFileName;}
	public String getExpProbModelFile() {return expProbModelFileName;}
	public String getHeadCatProbModelFile() {return headCatProbModelFileName;}
	public String getSisterCatProbModelFile() {return sisterCatProbModelFileName;}
	public String getSisterLexCatProbModelFile() {return sisterLexCatProbModelFileName;}
	public String getHeadMapFile() {return headMapFileName;}
	public double getAlpha() {return alpha;}
	public double getBackoffProb() {return backoffProb;}
	public int getSmoothCutOff()    {return smoothCutOff;}
	public double getSmoothLambda() {return smoothLambda;}
	public boolean checkCCG() {return ccgCheck;}
	public String getModelType(){return modelType;}
}